from __future__ import annotations

import copy
import datetime
import re
import zoneinfo
from typing import Any

import pytest
from hypothesis import assume, given
from hypothesis.strategies import composite, integers, none, one_of, sampled_from

from xocto import localtime, ranges


@composite
def valid_integer_range(draw):
    boundaries = draw(sampled_from(ranges.RangeBoundaries))
    # Restrict the size of the intervals since that will always be enough to find logic issues
    if boundaries in [
        ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE,
        ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE,
    ]:
        start = draw(integers(min_value=-5, max_value=5))
    else:
        start = draw(one_of(none(), integers(min_value=-5, max_value=5)))

    if start is None:
        min_end = -5
    elif boundaries == ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE:
        min_end = start
    else:
        min_end = start + 1
    # Give the end a bit more space so we can create intervals if start = 5
    if boundaries in [
        ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE,
        ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE,
    ]:
        end = draw(integers(min_end, 6))
    else:
        end = draw(one_of(none(), integers(min_end, 6)))

    return ranges.Range(start, end, boundaries=boundaries)


class TestRange:
    def test_creation(self):
        assert ranges.Range(
            0, 2, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE
        ) == ranges.Range(0, 2, boundaries="(]")

    def test_is_immutable(self):
        r = ranges.Range(0, 2, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE)
        with pytest.raises(AttributeError, match="Can't set attributes"):
            r.start = 1

    def test_additional_attributes_cant_be_created(self):
        r = ranges.Range(0, 2, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE)
        with pytest.raises(
            AttributeError, match="'Range' object has no attribute 'something_else'"
        ):
            r.something_else = 1

    def test_does_not_have_instance_dictionary(self):
        r = ranges.Range(0, 2, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE)
        assert not hasattr(r, "__dict__")

    @pytest.mark.parametrize(
        "start,end,boundaries",
        [
            (1, 0, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE),
            (0, 0, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE),
            (0, 0, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE),
            (0, 0, ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE),
            (None, 0, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE),
            (None, 0, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE),
            (None, None, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE),
            (0, None, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE),
            (0, None, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE),
            (None, None, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE),
        ],
    )
    def test_validation(self, start, end, boundaries):
        with pytest.raises(ValueError):
            ranges.Range(start, end, boundaries=boundaries)

    @pytest.mark.parametrize(
        "lower,higher",
        [
            ("[0,2)", "[1,3)"),
            ("[0,4)", "[1,3)"),
            ("[0,2)", "[0,4)"),
            ("[0,2)", "[0,None)"),
            ("(None,5)", "[0,4)"),
            ("[0,2)", "[0,2]"),
            ("[0,2]", "(0,2)"),
            ("(0,2)", "(0,2]"),
        ],
    )
    def test_range_comparison(self, lower, higher):
        lower_range = _range_from_string(lower)
        higher_range = _range_from_string(higher)

        assert higher_range > lower_range
        assert not (higher_range < lower_range)
        assert higher_range >= lower_range
        assert lower_range < higher_range
        assert lower_range <= higher_range

    @pytest.mark.parametrize(
        "start,end,boundaries,item,expected",
        [
            # Basic inclusion
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE, 0, True),
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE, 1, True),
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE, 2, True),
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE, 3, False),
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE, -1, False),
            # Unset bounds <=> Infinity
            (None, 2, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE, -1, True),
            (None, 2, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE, 3, False),
            (0, None, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE, -1, False),
            (0, None, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE, 3, True),
            (None, None, ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE, 3, True),
            (None, None, ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE, -1, True),
            # Non-integer comparable types
            (
                datetime.date(2020, 1, 1),
                datetime.date(2020, 1, 3),
                ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE,
                datetime.date(2020, 1, 2),
                True,
            ),
            (
                datetime.date(2020, 1, 1),
                datetime.date(2020, 1, 3),
                ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE,
                datetime.date(2019, 1, 1),
                False,
            ),
            # Different boundariess
            (0, 2, ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE, 0, False),
            (0, 2, ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE, 2, False),
            (0, 2, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE, 0, False),
            (0, 2, ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE, 2, True),
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE, 0, True),
            (0, 2, ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE, 2, False),
        ],
    )
    def test_contains(self, start, end, boundaries, item, expected):
        result = item in ranges.Range(start, end, boundaries=boundaries)
        assert result == expected

    def test_default_boundaries(self):
        subject = ranges.Range(0, 2)

        result = subject.boundaries

        assert result == ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE

    @given(valid_integer_range(), valid_integer_range())
    def test_range_is_disjoint(
        self, a: ranges.Range[Any], b: ranges.Range[Any]
    ) -> None:
        assert a.intersection(b) is not None or a.is_disjoint(b)

    @pytest.mark.parametrize(
        "a_str,b_str,expected_str",
        [
            # Finite intersections
            ("[0,2)", "[1,3)", "[1,2)"),
            ("[0,2]", "[1,3)", "[1,2]"),
            # Interactions with unbounded ranges
            ("[0,None)", "[1,3)", "[1,3)"),
            ("[0,None)", "[2,None)", "[2,None)"),
            ("[0,3)", "[2,None)", "[2,3)"),
            ("(None,3)", "[0,4)", "[0,3)"),
            ("(None,3)", "[0,None)", "[0,3)"),
            # Disjoint ranges
            ("[0,2)", "[2,4)", None),
            ("[0,2)", "[3,4)", None),
            ("(None,2)", "[2,None)", None),
            ("(None,2)", "[3,None)", None),
            # Other range types
            ("(0,2)", "(1,3)", "(1,2)"),
            ("(0,2)", "(2,4)", None),
            ("(0,2]", "(1,3]", "(1,2]"),
            ("(0,2]", "(2,4]", None),
            ("[0,2]", "[1,3]", "[1,2]"),
            ("[0,2]", "[2,4]", "[2,2]"),
            # Mixed range types
            ("(0,2)", "[1,3)", "[1,2)"),
            ("(0,2)", "[-1,1)", "(0,1)"),
            ("(0,3)", "[1,2)", "[1,2)"),
            ("(0,2)", "[2,3)", None),
            ("[0,2]", "(0,2)", "(0,2)"),
            ("(0,2]", "[2,4)", "[2,2]"),
        ],
    )
    def test_intersection(self, a_str, b_str, expected_str):
        a = _range_from_string(a_str)
        b = _range_from_string(b_str)
        if expected_str is not None:
            expected = _range_from_string(expected_str)
        else:
            expected = None

        result = a & b

        assert result == expected

    @pytest.mark.parametrize(
        "a_str,b_str,expected_str",
        [
            # Finite intersections
            ("[0,2)", "[1,3)", "[0,3)"),
            ("[1,3)", "[0,2)", "[0,3)"),
            # Interactions with unbounded ranges
            ("[1,None)", "[0,3)", "[0,None)"),
            ("[0,None)", "[2,None)", "[0,None)"),
            ("(None,3)", "[0,4)", "(None,4)"),
            ("(None,3)", "[0,None)", "(None,None)"),
            # Adjacent ranges
            ("[0,2)", "[2,4)", "[0,4)"),
            ("[0,2]", "(2,4)", "[0,4)"),
            # Disjoint ranges
            ("[0,2)", "[3,4)", None),
            ("(None,2)", "[3,None)", None),
            # Other range types
            ("(0,2)", "(1,3)", "(0,3)"),
            ("(0,2)", "(2,4)", None),
            ("(0,2]", "(1,3]", "(0,3]"),
            ("(0,2]", "(2,4]", "(0,4]"),
            ("[0,2]", "[1,3]", "[0,3]"),
            ("[0,2]", "[2,4]", "[0,4]"),
            # Mixed range types
            ("(0,2)", "[0,2)", "[0,2)"),
            ("(0,2)", "[2,3)", "(0,3)"),
            ("[0,2]", "(0,2)", "[0,2]"),
            ("[0,2]", "(0,3)", "[0,3)"),
            ("(0,2]", "[2,4)", "(0,4)"),
        ],
    )
    def test_range_union(self, a_str, b_str, expected_str):
        a = _range_from_string(a_str)
        b = _range_from_string(b_str)
        if expected_str is not None:
            expected = _range_from_string(expected_str)
        else:
            expected = None

        result = a | b

        assert result == expected

    @given(valid_integer_range(), valid_integer_range())
    def test_union_and_intersection_are_commutative(
        self, a: ranges.Range[Any], b: ranges.Range[Any]
    ) -> None:
        assert a | b == b | a
        assert a & b == b & a

    @given(valid_integer_range(), valid_integer_range())
    def test_union_and_intersection_are_idempotent(
        self, a: ranges.Range[Any], b: ranges.Range[Any]
    ) -> None:
        union = a | b
        assume(union is not None)
        assert union is not None
        assert union & a == a
        assert union & b == b

    @given(valid_integer_range(), valid_integer_range())
    def test_range_difference_and_intersection_form_partition(
        self, a: ranges.Range[Any], b: ranges.Range[Any]
    ) -> None:
        a_difference = a - b
        b_difference = b - a
        intersection = a & b

        assume(a_difference is not None or b_difference is not None)

        if intersection is None:
            assert a_difference == a
            assert b_difference == b
        else:
            if a_difference is not None:
                if isinstance(a_difference, ranges.RangeSet):
                    # a contains b
                    assert b_difference is None
                    assert a_difference.is_disjoint(ranges.RangeSet([intersection]))
                    assert a_difference | ranges.RangeSet(
                        [intersection]
                    ) == ranges.RangeSet([a])
                else:
                    assert a_difference.is_disjoint(intersection)
                    assert a_difference | intersection == a

            if b_difference is not None:
                if isinstance(b_difference, ranges.RangeSet):
                    # b contains a
                    assert a_difference is None
                    assert b_difference.is_disjoint(ranges.RangeSet([intersection]))
                    assert b_difference | ranges.RangeSet(
                        [intersection]
                    ) == ranges.RangeSet([b])
                else:
                    assert b_difference.is_disjoint(intersection)
                    assert b_difference | intersection == b

            if a_difference is not None and b_difference is not None:
                # Neither interval contains the other
                assert isinstance(a_difference, ranges.Range) and isinstance(
                    b_difference, ranges.Range
                )
                assert a_difference & b_difference is None
                assert a_difference.is_disjoint(b_difference)
                assert b_difference.is_disjoint(a_difference)
                # Ignore types here as structuring this to appease mypy would make it v ugly.
                assert (a_difference | intersection | b_difference) == (a | b)  # type: ignore[operator]

    class TestCopy:
        def test_range_copy(self):
            r1 = ranges.Range(1, 2)
            r2 = copy.copy(r1)
            assert r1 == r2

        def test_range_deepcopy(self):
            r1 = ranges.Range(1, 2)
            r2 = copy.deepcopy(r1)
            assert r1 == r2

        @pytest.mark.parametrize(
            "obj",
            [
                ranges.Range(1, 2),
                ranges.FiniteDateRange(
                    datetime.date(2000, 1, 1), datetime.date(2000, 1, 2)
                ),
                ranges.FiniteDatetimeRange(
                    datetime.datetime(2000, 1, 1), datetime.datetime(2000, 1, 2)
                ),
                ranges.HalfFiniteRange(1, 2),
            ],
            ids=("range", "date_range", "datetime_range", "half_finite_range"),
        )
        def test_copies(self, obj):
            assert obj == copy.copy(obj) == copy.deepcopy(obj)


class TestRangeSet:
    @pytest.mark.parametrize(
        "rangeset,expected_string",
        [
            # Empty RangeSets
            (ranges.RangeSet(), "{}"),
            (ranges.RangeSet([]), "{}"),
            # Single item sets
            (ranges.RangeSet([ranges.Range(0, 2)]), "{[0,2)}"),
            (ranges.RangeSet([ranges.Range(0, 3, boundaries="()")]), "{(0,3)}"),
            (ranges.RangeSet([ranges.Range(0, None)]), "{[0,None)}"),
            (
                ranges.RangeSet(
                    [
                        ranges.Range(
                            datetime.date(2020, 1, 1),
                            datetime.date(2020, 1, 3),
                        )
                    ]
                ),
                "{[2020-01-01,2020-01-03)}",
            ),
            # Multiple disjoint items
            (
                ranges.RangeSet([ranges.Range(0, 2), ranges.Range(3, 5)]),
                "{[0,2), [3,5)}",
            ),
            (
                ranges.RangeSet([ranges.Range(3, 5), ranges.Range(0, 2)]),
                "{[0,2), [3,5)}",
            ),
            # Overlapping items
            (ranges.RangeSet([ranges.Range(0, 2), ranges.Range(1, 3)]), "{[0,3)}"),
            (ranges.RangeSet([ranges.Range(1, 3), ranges.Range(0, 2)]), "{[0,3)}"),
        ],
    )
    def test_rangeset_construction(
        self, rangeset: ranges.RangeSet[Any], expected_string: str
    ) -> None:
        assert str(rangeset) == expected_string

    @given(valid_integer_range(), valid_integer_range())
    def test_rangeset_addition(
        self, a: ranges.Range[Any], b: ranges.Range[Any]
    ) -> None:
        a_set = ranges.RangeSet([a])
        b_set = ranges.RangeSet([b])

        a_set.add(b)
        b_set.add(a)

        assert a_set == b_set == ranges.RangeSet([a, b])

    @pytest.mark.parametrize(
        "rangeset,item,expected_result",
        [
            # Exact match
            (ranges.RangeSet([ranges.Range(0, 5)]), ranges.Range(0, 5), True),
            # Contained match
            (ranges.RangeSet([ranges.Range(0, 5)]), ranges.Range(1, 3), True),
            # Partial match
            (ranges.RangeSet([ranges.Range(0, 5)]), ranges.Range(1, 6), False),
            # Partial match
            (
                ranges.RangeSet([ranges.Range(0, 2), ranges.Range(3, 7)]),
                ranges.Range(1, 6),
                False,
            ),
        ],
    )
    def test_rangeset_contains_range(self, rangeset, item, expected_result):
        assert (item in rangeset) == expected_result

    @pytest.mark.parametrize(
        "item,expected_result",
        [
            (-1, False),
            (0, True),
            (1, True),
            (2, True),
            (3, False),
            (4, True),
            (5, False),
        ],
    )
    def test_rangeset_contains_comparable_item(self, item, expected_result):
        rangeset = _rangeset_from_string("{[0,2], [4,5)}")
        assert (item in rangeset) == expected_result

    @pytest.mark.parametrize(
        "rangeset_str, expected_result_str",
        [
            # Gaps
            ("{[0,1), [2,3), [4,None)}", "{(None,0), [1,2), [3,4)}"),
            # Alternative boundaries
            ("{[0,1], [2,3]}", "{(None,0), (1,2), (3,None)}"),
            ("{(0,1), (2,3)}", "{(None,0], [1,2], [3,None)}"),
            # No gaps
            ("{}", "{(None,None)}"),
            ("{[0,1)}", "{(None,0), [1,None)}"),
            ("{[0,None), [2,3), [4,5)}", "{(None,0)}"),
            ("{[0,1), [2,3), (None,5]}", "{(5,None)}"),
        ],
    )
    def test_rangeset_complement(self, rangeset_str, expected_result_str):
        rangeset = _rangeset_from_string(rangeset_str)
        assert str(-rangeset) == expected_result_str

    @pytest.mark.parametrize(
        "rangeset_str, other_rangeset_str, expected_result_str",
        [
            # Single range rangesets, no contained difference
            ("{[0,100)}", "{[0,200)}", "{}"),
            # Multi range rangeset, some contained differences
            ("{[0,100)}", "{[0,10), [50,60)}", "{[10,50), [60,100)}"),
            # Multi range rangeset, no differences
            ("{[0,10), [50,60)}", "{[0,100)}", "{}"),
            # Multi range rangeset, entirely different
            ("{[20,30), [50,60)}", "{[0,10), [40,50)}", "{[20,30), [50,60)}"),
            # Multi range rangeset, unbounded
            ("{[20,30), [50,None)}", "{[0,10), [50,60)}", "{[20,30), [60,None)}"),
            # Multi range rangeset, numerous leading ranges
            (
                "{(None,5], [10,15), [20,30), [50,None)}",
                "{[25,40), [50,60)}",
                "{(None,5], [10,15), [20,25), [60,None)}",
            ),
            # Alternate boundaries
            ("{[0,100]}", "{(0,100)}", "{[0,0], [100,100]}"),  # INC_INC, EXC_EXC
            ("{(0,100)}", "{(0,100)}", "{}"),  # EXC_EXC, EXC_EXC
            ("{(0,100)}", "{[0,100]}", "{}"),  # EXC_EXC, INC_INC
            ("{[0,100]}", "{[0,100]}", "{}"),  # INC_INC, INC_INC
            ("{[0,100)}", "{[0,100]}", "{}"),  # INC_EXC, INC_INC
            ("{[0,100]}", "{[0,100)}", "{[100,100]}"),  # INC_INC, INC_EXC
            ("{[0,100]}", "{(0,100]}", "{[0,0]}"),  # INC_INC, EXC_INC
            ("{(0,100]}", "{[0,100)}", "{[100,100]}"),  # EXC_INC, INC_EXC
            ("{[0,100)}", "{(0,100]}", "{[0,0]}"),  # INC_EXC, EXC_INC
            ("{[0,100)}", "{[0,100)}", "{}"),  # INC_EXC, INC_EXC
            ("{[0,101)}", "{[0,100]}", "{(100,101)}"),  # INC_EXC + 1, INC_INC
            ("{[0,101]}", "{[0,100]}", "{(100,101]}"),  # INC_INC + 1, INC_INC
            (
                "{[0,1], [2,5], [7,9]}",
                "{[0,1), (3,7), [9,10]}",
                "{[1,1], [2,3], [7,9)}",
            ),
        ],
    )
    def test_rangeset_difference(
        self, rangeset_str, other_rangeset_str, expected_result_str
    ):
        rangeset = _rangeset_from_string(rangeset_str)
        other_rangeset = _rangeset_from_string(other_rangeset_str)
        assert str(rangeset - other_rangeset) == expected_result_str


class TestHalfFiniteRange:
    def test_creation_uses_expected_defaults(self):
        r = ranges.HalfFiniteRange(0)
        assert 0 == r.start
        assert r.end is None
        assert ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE == r.boundaries

    def test_creation(self):
        r = ranges.HalfFiniteRange(1, 2)
        assert 1 == r.start
        assert 2 == r.end

    def test_does_not_have_instance_dictionary(self):
        r = ranges.HalfFiniteRange(0, 2)
        assert not hasattr(r, "__dict__")


ONE_DAY = datetime.timedelta(days=1)


class TestBreakPeriodsOnTimestamp:
    @pytest.fixture
    def period(self):
        return ranges.FiniteDatetimeRange(
            datetime.datetime(2021, 1, 1),
            datetime.datetime(2021, 2, 1),
        )

    def test_no_timestamps_returns_period(self, period):
        assert ranges.get_finite_datetime_ranges_from_timestamps(period, []) == [period]

    def test_timestamps_outside_period_ignored(self, period):
        assert ranges.get_finite_datetime_ranges_from_timestamps(
            period,
            [
                period.start - ONE_DAY,
                period.end + ONE_DAY,
            ],
        ) == [period]

    def test_timestamps_at_ends_ignored(self, period):
        assert ranges.get_finite_datetime_ranges_from_timestamps(
            period,
            [
                period.start,
                period.end,
            ],
        ) == [period]

    def test_timestamps_inside_period_cause_period_to_split(self, period):
        splits = [
            period.start + ONE_DAY,
            period.end - ONE_DAY,
        ]
        assert ranges.get_finite_datetime_ranges_from_timestamps(period, splits) == [
            ranges.FiniteDatetimeRange(period.start, splits[0]),
            ranges.FiniteDatetimeRange(splits[0], splits[1]),
            ranges.FiniteDatetimeRange(splits[1], period.end),
        ]

    def test_timestamps_deduplicated(self, period):
        splits = [
            period.start + ONE_DAY,
            period.start + ONE_DAY,
            period.end - ONE_DAY,
            period.end - ONE_DAY,
        ]
        assert ranges.get_finite_datetime_ranges_from_timestamps(period, splits) == [
            ranges.FiniteDatetimeRange(period.start, splits[0]),
            ranges.FiniteDatetimeRange(splits[0], splits[2]),
            ranges.FiniteDatetimeRange(splits[2], period.end),
        ]

    def test_timestamps_sorted(self, period):
        splits = [
            period.end - ONE_DAY,
            period.start + ONE_DAY,
        ]
        assert ranges.get_finite_datetime_ranges_from_timestamps(period, splits) == [
            ranges.FiniteDatetimeRange(period.start, splits[1]),
            ranges.FiniteDatetimeRange(splits[1], splits[0]),
            ranges.FiniteDatetimeRange(splits[0], period.end),
        ]


class TestAnyOverlapping:
    def test_does_not_modify_ranges(self):
        # The implementation of `any_overlapping` relies on sorting.
        # Let's make sure that the ranges passed in are unchanged.
        ranges_ = [
            ranges.Range(1, 2),
            ranges.Range(0, 1),
        ]
        ranges_copy = ranges_.copy()
        assert not ranges.any_overlapping(ranges_)
        assert ranges_ == ranges_copy

    @pytest.mark.parametrize(
        "ranges_",
        [
            [
                ranges.Range(0, 2),
                ranges.Range(1, 3),
            ],
            [
                ranges.Range(0, 2),
                ranges.Range(
                    4, 5
                ),  # Added this to ensure the overlapping ranges are not adjacent.
                ranges.Range(1, 3),
            ],
            [
                ranges.Range(
                    0, 2, boundaries=ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE
                ),
                ranges.Range(2, 4),
            ],
        ],
    )
    def test_returns_true_if_and_ranges_overlap(self, ranges_):
        assert ranges.any_overlapping(ranges_)
        assert ranges.any_overlapping(reversed(ranges_))

    @pytest.mark.parametrize(
        "ranges_",
        [
            [
                ranges.Range(0, 2),
                ranges.Range(2, 4),
            ],
            [
                ranges.Range(0, 2),
                ranges.Range(
                    2, 4, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE
                ),
            ],
        ],
    )
    def test_returns_false_if_no_ranges_overlap(self, ranges_):
        assert not ranges.any_overlapping(ranges_)
        assert not ranges.any_overlapping(reversed(ranges_))

    def test_returns_false_for_empty_set_of_ranges(self):
        assert not ranges.any_overlapping([])


class TestAnyGaps:
    def test_does_not_modify_ranges(self):
        # The implementation of `any_gaps` relies on sorting.
        # Let's make sure that the ranges passed in are unchanged.
        ranges_ = [
            ranges.Range(1, 2),
            ranges.Range(0, 1),
        ]
        ranges_copy = ranges_.copy()
        assert not ranges.any_gaps(ranges_)
        assert ranges_ == ranges_copy

    @pytest.mark.parametrize(
        "ranges_",
        [
            [
                ranges.Range(0, 1),
                ranges.Range(2, 3),
            ],
            [
                ranges.Range(
                    0, 1, boundaries=ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE
                ),
                ranges.Range(
                    1, 2, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE
                ),
            ],
            [
                ranges.Range(0, 2),
                ranges.Range(4, 6),
                ranges.Range(1, 3),
            ],
        ],
    )
    def test_returns_true_if_gaps(self, ranges_):
        assert ranges.any_gaps(ranges_)
        assert ranges.any_gaps(reversed(ranges_))

    @pytest.mark.parametrize(
        "ranges_",
        [
            [
                ranges.Range(0, 1),
                ranges.Range(1, 2),
            ],
            [
                ranges.Range(
                    0, 1, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE
                ),
                ranges.Range(
                    1, 2, boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE
                ),
            ],
            [
                ranges.Range(0, 2),
                ranges.Range(1, 3),
            ],
            [
                ranges.Range(0, 3),
                ranges.Range(1, 2),
            ],
            [
                ranges.Range(0, 5),
                ranges.Range(1, 2),
                ranges.Range(3, 4),
            ],
            [
                ranges.Range(0, 2),
                ranges.Range(4, 6),
                ranges.Range(2, 4),
            ],
        ],
    )
    def test_returns_false_if_no_gaps(self, ranges_):
        assert not ranges.any_gaps(ranges_)
        assert not ranges.any_gaps(reversed(ranges_))

    def test_returns_false_for_empty_set_of_ranges(self):
        assert not ranges.any_gaps([])


class TestFiniteDateRange:
    """
    Test class for methods specific to the the FiniteDateRange subclass.
    """

    def test_days_property_has_correct_value(self):
        """
        As FiniteDateRange boundaries are double inclusive, the days
        property should include the start and end dates in the count.
        """
        range = ranges.FiniteDateRange(
            start=datetime.date(2000, 1, 1),
            end=datetime.date(2000, 1, 2),
        )
        assert range.days == 2

    class TestIntersection:
        def test_overlapping_ranges_will_intersect(self):
            ranges_ = [
                ranges.FiniteDateRange(
                    start=datetime.date(2000, 1, 1),
                    end=datetime.date(2000, 1, 3),
                ),
                ranges.FiniteDateRange(
                    start=datetime.date(2000, 1, 3),
                    end=datetime.date(2000, 1, 5),
                ),
            ]
            assert ranges_[0] & ranges_[1] == ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 3),
            )

        def test_adjacent_ranges_wont_intersect(self):
            ranges_ = [
                ranges.FiniteDateRange(
                    start=datetime.date(2000, 1, 1),
                    end=datetime.date(2000, 1, 2),
                ),
                ranges.FiniteDateRange(
                    start=datetime.date(2000, 1, 3),
                    end=datetime.date(2000, 1, 4),
                ),
            ]
            assert ranges_[0] & ranges_[1] is None

    class TestIsDisjoint:
        @pytest.mark.parametrize(
            "other",
            [
                ranges.Range(
                    start=datetime.date(2000, 1, 1),
                    end=datetime.date(2000, 1, 2),
                    boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE,
                ),
                ranges.Range(
                    start=datetime.date(2000, 1, 5),
                    end=datetime.date(2000, 1, 6),
                    boundaries=ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE,
                ),
            ],
        )
        def test_identifies_ranges_that_are_not_disjoint(self, other):
            range = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 4),
            )
            assert range.is_disjoint(other) is False

        def test_handles_edge_of_time(self):
            # This test is fairly coupled to knowledge of the is_disjoint implementation here.
            range = ranges.FiniteDateRange(
                start=datetime.date.min,
                end=datetime.date.max,
            )
            # We're not particularly interested in this assertion, we just want to be sure this
            # doesn't error.
            assert range.is_disjoint(range) is False

        @pytest.mark.parametrize(
            "other",
            [
                ranges.Range(
                    start=datetime.date(2000, 1, 1),
                    end=datetime.date(2000, 1, 2),
                    boundaries=ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE,
                ),
                ranges.Range(
                    start=datetime.date(2000, 1, 5),
                    end=datetime.date(2000, 1, 6),
                    boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE,
                ),
            ],
        )
        def test_identifies_ranges_that_are_disjoint(self, other):
            range = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 4),
            )
            assert range.is_disjoint(other) is True

    class TestUnion:
        @pytest.mark.parametrize(
            "range, other",
            [
                (
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 1),
                        end=datetime.date(2000, 1, 2),
                    ),
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 3),
                        end=datetime.date(2000, 1, 4),
                    ),
                ),
                (
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 3),
                        end=datetime.date(2000, 1, 4),
                    ),
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 1),
                        end=datetime.date(2000, 1, 2),
                    ),
                ),
            ],
        )
        def test_handles_adjacent_ranges(self, range, other):
            """
            As FiniteDateRange boundaries are double inclusive, when two ranges
            end/start on consecutive days they effectively cover a fully contiguous
            period of time and should make a union.
            """
            union = range | other
            assert union == ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 1),
                end=datetime.date(2000, 1, 4),
            )

        def test_handles_many_adjacent_ranges(self):
            """
            As FiniteDateRange boundaries are double inclusive, when two ranges
            end/start on consecutive days they effectively cover a fully contiguous
            period of time and should make a union.

            This should apply with an arbitrary number of adjacent ranges.
            """
            r1 = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 1),
                end=datetime.date(2000, 1, 2),
            )
            r2 = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 5),
            )
            r3 = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 6),
                end=datetime.date(2000, 1, 8),
            )

            union = r1 | r2 | r3
            assert union == ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 1),
                end=datetime.date(2000, 1, 8),
            )

        def test_range_set_over_many_adjacent_ranges(self):
            """
            Multiple unions should be approximately equivalent to a RangeSet
            """
            r1 = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 1),
                end=datetime.date(2000, 1, 2),
            )
            r2 = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 5),
            )
            r3 = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 6),
                end=datetime.date(2000, 1, 8),
            )

            range_set = ranges.RangeSet([r1, r2, r3])
            assert (
                ranges.FiniteDateRange(
                    start=datetime.date(2000, 1, 1),
                    end=datetime.date(2000, 1, 8),
                )
                in range_set
            )

        @pytest.mark.parametrize(
            "range, other",
            [
                (
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 1),
                        end=datetime.date(2000, 1, 2),
                    ),
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 4),
                        end=datetime.date(2000, 1, 5),
                    ),
                ),
                (
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 4),
                        end=datetime.date(2000, 1, 5),
                    ),
                    ranges.FiniteDateRange(
                        start=datetime.date(2000, 1, 1),
                        end=datetime.date(2000, 1, 2),
                    ),
                ),
            ],
        )
        def test_handles_disjoint_ranges(self, range, other):
            """
            Disjoint ranges (those with a full clear day between them) should not make
            a union.
            """
            union = range | other
            assert union is None

        @pytest.mark.parametrize(
            "other",
            [
                pytest.param(
                    # This is a complex case, because the behaviour of union orders
                    # the ranges first the union method is called on this param Range
                    # not the FiniteDateRange defined in the test body. The test case
                    # still has value though as it does confirm that we get the desired
                    # behaviour, even if it is not with the expected method being
                    # called.
                    ranges.Range(
                        start=datetime.date(2000, 1, 1),
                        end=datetime.date(2000, 1, 2),
                        boundaries=ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE,
                    ),
                    id="other_excludes_adjacent_end_boundary",
                ),
                pytest.param(
                    ranges.Range(
                        start=datetime.date(2000, 1, 5),
                        end=datetime.date(2000, 1, 6),
                        boundaries=ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE,
                    ),
                    id="other_excludes_adjacent_start_boundary",
                ),
            ],
        )
        def test_respects_other_range_boundaries(self, other):
            """
            Disjoint ranges (those with a full clear day between them) should not make
            a union.
            """
            range = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 4),
            )
            union = range | other
            assert union is None

        def test_doesnt_extend_union(self):
            """
            A union of ranges should be longer than the sum of it's parts.
            """
            # This is a weird test to include, it is added because this feels like an
            # obvious risk with the implementation I have used.
            range = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 4),
            )
            other = ranges.FiniteDateRange(
                start=datetime.date(2000, 1, 3),
                end=datetime.date(2000, 1, 4),
            )
            union = range | other
            assert union == range

    class TestIn:
        def test_finite_range(self):
            subject = ranges.FiniteRange(1, 4)

            assert 3 in subject


class TestFiniteDatetimeRange:
    class TestUnion:
        def test_union_of_touching_ranges(self):
            range = ranges.FiniteDatetimeRange(
                start=datetime.datetime(2000, 1, 1),
                end=datetime.datetime(2000, 1, 2),
            )
            other = ranges.FiniteDatetimeRange(
                start=datetime.datetime(2000, 1, 2),
                end=datetime.datetime(2000, 1, 3),
            )

            union = range | other

            assert union == ranges.FiniteDatetimeRange(
                start=datetime.datetime(2000, 1, 1),
                end=datetime.datetime(2000, 1, 3),
            )

        def test_union_of_disjoint_ranges(self):
            range = ranges.FiniteDateRange(
                start=datetime.datetime(2000, 1, 1),
                end=datetime.datetime(2000, 1, 2),
            )
            other = ranges.FiniteDatetimeRange(
                start=datetime.datetime(2020, 1, 1),
                end=datetime.datetime(2020, 1, 2),
            )

            assert range | other is None

        def test_union_of_overlapping_ranges(self):
            range = ranges.FiniteDatetimeRange(
                start=datetime.datetime(2000, 1, 1),
                end=datetime.datetime(2000, 1, 3),
            )
            other = ranges.FiniteDatetimeRange(
                start=datetime.datetime(2000, 1, 2),
                end=datetime.datetime(2000, 1, 4),
            )

            union = range | other

            assert union == ranges.FiniteDatetimeRange(
                start=datetime.datetime(2000, 1, 1),
                end=datetime.datetime(2000, 1, 4),
            )

    class TestLocalize:
        def test_converts_timezone(self):
            # Create a datetime range in Sydney, which is
            # 7 hours ahead of Dubai (target timezone).
            source_tz = zoneinfo.ZoneInfo("Australia/Sydney")  # GMT+11
            target_tz = zoneinfo.ZoneInfo("Asia/Dubai")  # GMT+4

            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 1, 1, hour=7, tzinfo=source_tz),
                datetime.datetime(2020, 1, 10, hour=7, tzinfo=source_tz),
            )

            assert dt_range.localize(target_tz) == ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 1, 1, tzinfo=target_tz),
                datetime.datetime(2020, 1, 10, tzinfo=target_tz),
            )

        def test_errors_converting_over_dst_gain_hour(self):
            utc_tz = zoneinfo.ZoneInfo("UTC")
            london_tz = zoneinfo.ZoneInfo("Europe/London")

            # Create a range in London over the hour that is "gained"
            # when Daylight Savings Time (DST) starts - at 1AM.
            #
            # Note: this is allowed by datetime but not a realistic
            # example - "1AM" here doesn't actually exist in GBR.
            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 3, 29, hour=1, tzinfo=london_tz),
                datetime.datetime(2020, 3, 29, hour=2, tzinfo=london_tz),
            )

            # Converting to UTC should error due to the period being
            # empty: removing the "fake hour" means 2AM => 1AM.
            with pytest.raises(ValueError):
                assert dt_range.localize(utc_tz)

        def test_errors_converting_over_dst_loss_hour(self):
            utc_tz = zoneinfo.ZoneInfo("UTC")
            london_tz = zoneinfo.ZoneInfo("Europe/London")

            # Create a range in UTC over the hour before Daylight Savings
            # Time (DST) ends - at 2AM.
            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 10, 25, hour=0, tzinfo=utc_tz),
                datetime.datetime(2020, 10, 25, hour=1, tzinfo=utc_tz),
            )

            # Converting to London timezone should error due to the period
            # being empty: both times map to 1AM.
            with pytest.raises(ValueError):
                assert dt_range.localize(london_tz)

        def test_errors_if_naive(self):
            tz = zoneinfo.ZoneInfo("Europe/London")

            with pytest.raises(ValueError) as exc_info:
                ranges.FiniteDatetimeRange(
                    datetime.datetime(2020, 1, 1),
                    datetime.datetime(2020, 1, 10),
                ).localize(tz)

            assert "naive" in str(exc_info.value)

    class TestAsDateRange:
        def test_returns_date_range(self):
            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 10),
            )

            assert dt_range.as_date_range() == ranges.FiniteDateRange(
                datetime.date(2020, 1, 1),
                datetime.date(2020, 1, 9),
            )

        def test_errors_if_different_timezones(self):
            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 1, 1, tzinfo=zoneinfo.ZoneInfo("Asia/Dubai")),
                datetime.datetime(
                    2020, 1, 10, tzinfo=zoneinfo.ZoneInfo("Australia/Sydney")
                ),
            )

            with pytest.raises(ValueError) as exc_info:
                dt_range.as_date_range()

            assert "Start and end in different timezones" in str(exc_info.value)

        def test_errors_if_start_not_midnight(self):
            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 1, 1, hour=1),
                datetime.datetime(2020, 1, 10),
            )

            with pytest.raises(ValueError) as exc_info:
                dt_range.as_date_range()

            assert "Start of range is not midnight-aligned" in str(exc_info.value)

        def test_errors_if_end_not_midnight(self):
            dt_range = ranges.FiniteDatetimeRange(
                datetime.datetime(2020, 1, 1),
                datetime.datetime(2020, 1, 10, hour=1),
            )

            with pytest.raises(ValueError) as exc_info:
                dt_range.as_date_range()

            assert "End of range is not midnight-aligned" in str(exc_info.value)


class TestAsFiniteDatetimePeriods:
    def test_converts(self):
        actually_finite_range = ranges.DatetimeRange(
            datetime.datetime(2020, 1, 1),
            datetime.datetime(2020, 1, 5),
        )

        results = list(
            ranges.as_finite_datetime_periods(
                [actually_finite_range],
            )
        )

        assert results == [
            ranges.FiniteDatetimeRange(
                actually_finite_range.start,
                actually_finite_range.end,
            )
        ]

    def test_errors_if_infinite(self):
        with pytest.raises(ValueError) as exc_info:
            ranges.as_finite_datetime_periods(
                [
                    ranges.DatetimeRange(datetime.datetime(2020, 1, 1), None),
                ],
            )

        assert "Period is not finite at start or end or both" in str(exc_info.value)


class TestIterateOverMonths:
    @pytest.mark.parametrize(
        "row",
        [
            {
                "period": ranges.FiniteDatetimeRange(
                    datetime.datetime(2023, 6, 12, tzinfo=localtime.UTC),
                    datetime.datetime(2023, 6, 26, tzinfo=localtime.UTC),
                ),
                "expected": [
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2023, 6, 12, tzinfo=localtime.UTC),
                        datetime.datetime(2023, 6, 26, tzinfo=localtime.UTC),
                    )
                ],
                "id": "within-one-month",
            },
            {
                "period": ranges.FiniteDatetimeRange(
                    datetime.datetime(2021, 5, 12, tzinfo=localtime.UTC),
                    datetime.datetime(2021, 6, 26, tzinfo=localtime.UTC),
                ),
                "expected": [
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 5, 12, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 6, 1, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 6, 1, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 6, 26, tzinfo=localtime.UTC),
                    ),
                ],
                "id": "spanning-two-months",
            },
            {
                "period": ranges.FiniteDatetimeRange(
                    datetime.datetime(2021, 4, 12, tzinfo=localtime.UTC),
                    datetime.datetime(2021, 6, 26, tzinfo=localtime.UTC),
                ),
                "expected": [
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 4, 12, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 5, 1, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 5, 1, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 6, 1, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 6, 1, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 6, 26, tzinfo=localtime.UTC),
                    ),
                ],
                "id": "spanning-three-months",
            },
            {
                "period": ranges.FiniteDatetimeRange(
                    datetime.datetime(2021, 1, 31, 0, 0, tzinfo=localtime.UTC),
                    datetime.datetime(2022, 2, 1, 0, 0, tzinfo=localtime.UTC),
                ),
                "expected": [
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 1, 31, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 2, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 2, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 3, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 3, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 4, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 4, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 5, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 5, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 6, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 6, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 7, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 7, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 8, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 8, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 9, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 9, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 10, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 10, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 11, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 11, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2021, 12, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2021, 12, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2022, 1, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                    ranges.FiniteDatetimeRange(
                        datetime.datetime(2022, 1, 1, 0, 0, tzinfo=localtime.UTC),
                        datetime.datetime(2022, 2, 1, 0, 0, tzinfo=localtime.UTC),
                    ),
                ],
                "id": "spanning-thirteen-months",
            },
        ],
        ids=lambda row: row["id"],
    )
    def test_yields_correct_ranges(self, row):
        result = list(
            ranges.iterate_over_months(period=row["period"], tz=localtime.UTC)
        )

        assert result == row["expected"]


def _rangeset_from_string(rangeset_str: str) -> ranges.RangeSet[int]:
    """
    Convenience method to make test declarations clearer.

    Examples:
    {} => ranges.RangeSet()
    {[1,2]} => ranges.RangeSet(
        ranges.Range(1, 2, boundaries=ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE)
    )
    {[1,2], (3,None)} => ranges.RangeSet(
        ranges.Range(1, 2, boundaries=ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE),
        ranges.Range(3, None, boundaries=ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE)
    )
    """
    range_strs = re.findall(r"[\[\(][^\]\)]*[^\[\(][\]\)]", rangeset_str[1:-1])
    return ranges.RangeSet([_range_from_string(range_str) for range_str in range_strs])


def _range_from_string(range_str: str) -> ranges.Range[int]:
    """
    Convenience method to make test declarations clearer.

    Examples:
    [1,2] => ranges.Range(1, 2, boundaries=ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE)
    """
    left_bracket = range_str[0]
    right_bracket = range_str[-1]
    start_str, end_str = range_str[1:-1].split(",")

    boundaries = {
        ("[", "]"): ranges.RangeBoundaries.INCLUSIVE_INCLUSIVE,
        ("[", ")"): ranges.RangeBoundaries.INCLUSIVE_EXCLUSIVE,
        ("(", "]"): ranges.RangeBoundaries.EXCLUSIVE_INCLUSIVE,
        ("(", ")"): ranges.RangeBoundaries.EXCLUSIVE_EXCLUSIVE,
    }[left_bracket, right_bracket]

    if start_str == "None":
        start = None
    else:
        start = int(start_str)

    if end_str == "None":
        end = None
    else:
        end = int(end_str)

    return ranges.Range(start, end, boundaries=boundaries)



from __future__ import annotations

import calendar
import datetime as datetime_
import decimal
import zoneinfo
from typing import Generator, Sequence, Tuple

from dateutil import tz
from dateutil.relativedelta import relativedelta
from django.utils import timezone

from . import numbers, ranges


# Timezone aware datetime in the far future.
far_future = timezone.make_aware(datetime_.datetime.max - datetime_.timedelta(days=2))

# Timezone aware datetime in the far past.
far_past = timezone.make_aware(datetime_.datetime.min + datetime_.timedelta(days=2))

UTC = zoneinfo.ZoneInfo("UTC")
LONDON = zoneinfo.ZoneInfo("Europe/London")

ONE_DAY = datetime_.timedelta(days=1)
ONE_HOUR = datetime_.timedelta(hours=1)

MIDNIGHT_TIME = datetime_.time(0, 0)


def as_localtime(
    dt: datetime_.datetime, tz: datetime_.tzinfo | None = None
) -> datetime_.datetime:
    """
    Convert a tz aware datetime to localtime.

    Wrapper for the `django.utils.timezone` function, taking the same arguments.
    """
    return timezone.localtime(dt, timezone=tz)


def as_utc(dt: datetime_.datetime) -> datetime_.datetime:
    """
    Wrapper for normalizing a datetime aware object into UTC.
    """
    return as_localtime(dt, UTC)


def now() -> datetime_.datetime:
    """
    Return the current datetime in the local timezone.
    """
    return as_localtime(timezone.now())


def datetime(
    year: int,
    month: int,
    day: int,
    hour: int = 0,
    minute: int = 0,
    second: int = 0,
    microsecond: int = 0,
) -> datetime_.datetime:
    """
    Return a datetime in the local timezone.
    """
    dt = datetime_.datetime(year, month, day, hour, minute, second, microsecond)
    return timezone.make_aware(dt)


# Returning dates


def date(dt: datetime_.datetime, tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return the date of the given datetime in the given timezone, defaulting to local time.

    Note that `dt` must already be timezone-aware: it cannot be naive. It may be in UTC, for
    example, or already in current local time, in which case the effect will be the same as
    calling `.date()` on it. If `tz` is not provided, the function will return the date of `dt` in
    the local time zone.
    """
    if dt is None:
        # This is necessary as as_localtime is fine with being passed None, but we don't want to
        # accidentally do that here (since we have today as a separate method).
        raise TypeError("You must supply a datetime to localtime.date")

    return as_localtime(dt, tz=tz).date()


def today(tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return the current date in the provided timezone (or the local timezone if none is supplied).
    """
    return date(timezone.now(), tz)


def yesterday(tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return the previous date in the provided timezone (or the local timezone if none is supplied).
    """
    return days_in_the_past(1, tz)


def tomorrow(tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return the next date in the provided timezone (or the local timezone if none is supplied).
    """
    return days_in_the_future(1, tz)


def days_in_the_past(n: int, tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return n days before the current date (in the provided or local timezone).
    """
    return today(tz) - relativedelta(days=n)


def days_in_the_future(n: int, tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return n days after the current date (in the provided or local timezone).
    """
    return today(tz) + relativedelta(days=n)


def months_in_the_past(n: int, tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return n months before the current date (in the provided or local timezone).
    """
    return today(tz) - relativedelta(months=n)


def months_in_the_future(n: int, tz: datetime_.tzinfo | None = None) -> datetime_.date:
    """
    Return n months after the current date (in the provided or local timezone).
    """
    return today(tz) + relativedelta(months=n)


def date_of_day_before(
    dt: datetime_.datetime, tz: datetime_.tzinfo | None = None
) -> datetime_.date:
    """
    Return the date of the day before the datetime passed in.

    That is, find the local date of the datetime passed in, subtract one day from it, and return it.

    This is frequently useful when we are converting from the upper bound of a datetime range,
    which by convention we treat as exclusive, to the upper bound of a date range, which by
    convention we consider *inclusive*.

    Note that `dt` must already be timezone-aware: it cannot be naive. It may be in UTC, for
    example, or already in current local time, in which case the effect will be the same as
    calling `.date()` on it and subtracting a day. If `tz` is not provided, the function will
    return the date before `dt` in the local time zone.
    """
    return day_before(date(dt, tz))


def day_before(d: datetime_.date) -> datetime_.date:
    """
    Return the date of the day before the passed date.
    """
    return d - datetime_.timedelta(days=1)


def day_after(d: datetime_.date) -> datetime_.date:
    """
    Return the date of the day after the passed date.
    """
    return d + datetime_.timedelta(days=1)


# Returning datetimes


def seconds_in_the_future(
    n: int, dt: datetime_.datetime | None = None
) -> datetime_.datetime:
    """
    Return a datetime of the number of specifed seconds in the future.
    """
    if dt is None:
        dt = now()

    return dt + relativedelta(seconds=n)


def seconds_in_the_past(n: int) -> datetime_.datetime:
    """
    Return a datetime of the number of specifed seconds in the past.
    """
    return now() - relativedelta(seconds=n)


# Converting dates into datetimes


def midnight(
    date_or_datetime: datetime_.date | datetime_.datetime | None = None,
    tz: datetime_.tzinfo | None = None,
) -> datetime_.datetime:
    """
    Return a TZ-aware datetime for midnight of the passed date.

    If date is None then the current date in the passed timezone is used (if no timezone is
    supplied then we use the local timezone).
    """
    if date_or_datetime is None:
        date_: datetime_.date = today(tz)
    elif isinstance(date_or_datetime, datetime_.datetime):
        # Although this function is meant to be used on dates, we want to handle datetimes
        # properly as well, as these are frequently passed in as a way of 'truncating' them to
        # midnight on that date.
        if timezone.is_naive(date_or_datetime):
            # Default to localtime if no tz provided, as_localtime takes tz aware args.
            date_ = date(timezone.make_aware(date_or_datetime), tz)
        else:
            date_ = date(date_or_datetime, tz)
    else:
        date_ = date_or_datetime

    naive_midnight = datetime_.datetime.combine(date_, datetime_.datetime.min.time())
    return timezone.make_aware(naive_midnight, timezone=tz)


def next_midnight(
    date_or_datetime: datetime_.date | datetime_.datetime | None = None,
    tz: datetime_.tzinfo | None = None,
) -> datetime_.datetime:
    """
    Return the datetime for midnight of the following day to the date passed in.

    This is intuitively what people think of as midnight.
    """
    if date_or_datetime is None:
        date_: datetime_.date = today(tz)
    elif isinstance(date_or_datetime, datetime_.datetime):
        # Although this function is meant to be used on dates, we want to handle datetimes
        # properly as well
        if timezone.is_naive(date_or_datetime):
            # Default to localtime if no tz provided, as_localtime takes tz aware args
            date_ = date(timezone.make_aware(date_or_datetime), tz)
        else:
            date_ = date(date_or_datetime, tz)
    else:
        date_ = date_or_datetime

    return midnight(date_ + datetime_.timedelta(days=1), tz)


def midday(
    _date: datetime_.date | None = None, tz: datetime_.tzinfo | None = None
) -> datetime_.datetime:
    """
    Return a TZ-aware datetime for midday of the passed date.

    If date is None then the current date in the passed timezone is used (if no timezone is
    supplied then we use the local timezone).
    """
    if _date is None:
        _date = today(tz)

    return datetime_from_date(_date, hour=12, tz=tz)


def datetime_from_date(
    _date: datetime_.date, hour: int, tz: datetime_.tzinfo | None = None
) -> datetime_.datetime:
    """
    Return a TZ-aware datetime for the hour of the passed date.
    """
    naive_datetime = datetime_.datetime.combine(_date, datetime_.time(hour))
    return timezone.make_aware(naive_datetime, timezone=tz)


def datetime_from_epoch_timestamp(
    timestamp: int | float, tz: datetime_.tzinfo | None = None
) -> datetime_.datetime:
    """
    Return a TZ-aware datetime for the passed epoch timestamp.
    """
    naive_datetime_in_utc = datetime_.datetime.utcfromtimestamp(timestamp)
    utc_dt = timezone.make_aware(naive_datetime_in_utc, timezone=UTC)
    return timezone.localtime(utc_dt, timezone=tz)


def latest(
    _date: datetime_.date | None = None, tz: datetime_.tzinfo | None = None
) -> datetime_.datetime:
    """
    Return a TZ-aware datetime for the latest representable datetime of the passed date.

    If date is None then the current date in the local timezone is used.
    """
    if _date is None:
        _date = today()
    naive_midnight = datetime_.datetime.combine(_date, datetime_.datetime.max.time())
    return timezone.make_aware(naive_midnight, timezone=tz)


def combine(
    _date: datetime_.date, _time: datetime_.time, tz: datetime_.tzinfo
) -> datetime_.datetime:
    """
    Return a TZ-aware datetime obtained by combining the given date and time.

    It's a TZ-aware wrapper around datetime.datetime.combine.
    """
    combined_dt = datetime_.datetime.combine(_date, _time)
    return timezone.make_aware(combined_dt, timezone=tz)


# Converting dates into datetime pairs


def date_boundaries(
    _date: datetime_.date | None, tz: datetime_.tzinfo | None = None
) -> Tuple[datetime_.datetime, datetime_.datetime]:
    """
    Return a 2-tuple with the start and ending dt for the given date in the local timezone.

    Note, be careful of using this with __range ORM queries as such queries are INCLUSIVE on the
    boundaries.
    """
    return midnight(_date, tz), next_midnight(_date, tz)


def month_boundaries(
    month: int, year: int
) -> Tuple[datetime_.datetime, datetime_.datetime]:
    """
    Return the boundary datetimes of a given month.

    The boundary consists of a 2-tuple of: midnight of the first day of the month and midnight of the first day of the next month.

    For example:

        >>> month_boundaries(1, 2020)
        ((datetime.datetime(2020, 1, 1, 0, 0), datetime.datetime(2020, 2, 1, 0, 0))
    """
    start_date = datetime_.date(year, month, 1)
    end_date = start_date + relativedelta(months=1)
    return midnight(start_date), midnight(end_date)


def as_range(
    _date: datetime_.date | None, tz: datetime_.tzinfo | None = None
) -> Tuple[datetime_.datetime, datetime_.datetime]:
    """
    Return a 2-tuple of the min and max datetimes for the given date.

    This gives values that can be passed to the ORM __range filter.
    """
    return midnight(_date, tz), latest(_date, tz)


def make_aware_assuming_local(dt: datetime_.datetime) -> datetime_.datetime:
    """
    Just a wrapper for Django's method, which will takes a naive datetime, and makes it timezone
    aware, assuming the current timezone if none is passed (which it isn't from this wrapper
    function). It will also raise an exception if the passed datetime is already timezone-aware.
    """
    return timezone.make_aware(dt)


def make_aware_assuming_utc(dt: datetime_.datetime) -> datetime_.datetime:
    """
    Return a timezone-aware datetime (in UTC) given a naive datetime.
    """
    return timezone.make_aware(dt, timezone=UTC)


def is_utc(dt: datetime_.datetime) -> bool:
    """
    Test whether a given (timezone-aware) datetime is in UTC time or not.
    """
    return str(dt.tzinfo) == "UTC"


def is_local_time(dt: datetime_.datetime) -> bool:
    """
    Test whether a given (timezone-aware) datetime is in local time or not.
    """
    current_timezone = timezone.get_current_timezone()
    return dt.tzinfo == current_timezone


def within_date_range(
    first_date: datetime_.date, second_date: datetime_.date, days: int = 3
) -> bool:
    """
    Check if two dates are within a range from each other.
    """
    margin = datetime_.timedelta(days=days)
    return first_date - margin <= second_date <= first_date + margin


def quantise(
    dt: datetime_.datetime,
    timedelta: datetime_.timedelta,
    rounding: str = decimal.ROUND_HALF_EVEN,
) -> datetime_.datetime:
    """
    'Round' a datetime to the nearest interval given by the `timedelta` argument.

    For example:

        >>>> quantise(datetime.datetime(2020, 4, 1, 23, 21), timedelta(minutes=15))
        datetime.datetime(2020, 4, 1, 23, 30)

    """
    # We simply convert the datetime we want to quantise into a timestamp and use
    # `numbers.quantise()` to quantise it, with the 'seconds' from the timedelta argument as our
    # base.
    timedelta_seconds = timedelta.days * 24 * 60 * 60 + timedelta.seconds
    dt_as_timestamp = dt.timestamp()
    quantised_dt_timestamp = numbers.quantise(
        dt_as_timestamp, timedelta_seconds, rounding=rounding
    )
    quantised_dt = datetime_.datetime.fromtimestamp(
        quantised_dt_timestamp, tz=dt.tzinfo
    )
    return as_localtime(quantised_dt)


def nearest_half_hour(dt: datetime_.datetime) -> datetime_.datetime:
    """
    Return the nearest half-hour datetime to the passed datetime.
    """
    return quantise(dt, datetime_.timedelta(minutes=30))


def is_last_day_of_month(_date: datetime_.date) -> bool:
    """
    Return True if the given date is the last day of the month.
    """
    next_day = _date + datetime_.timedelta(days=1)
    if _date.month != next_day.month:
        return True
    return False


def start_of_month(dt: datetime_.datetime | None = None) -> datetime_.datetime:
    """
    Return the start datetime of the month for dt passed - or of current month.
    """
    if not dt:
        dt = now()
    return midnight(dt + relativedelta(day=1))


def end_of_month(dt: datetime_.datetime | None = None) -> datetime_.datetime:
    """
    Return the start datetime of the next month for dt passed - or of next month.
    """
    if not dt:
        dt = now()
    return midnight(dt + relativedelta(day=1, months=1))


def first_day_of_month(dt: datetime_.datetime | None = None) -> datetime_.date:
    """
    Return the start date of the month for dt passed - or of current month.
    """
    if dt is None:
        dt = now()
    return (dt + relativedelta(day=1)).date()


def last_day_of_month(dt: datetime_.datetime | None = None) -> datetime_.date:
    """
    Return the last date of the month for dt passed - or of current month.
    """
    if dt is None:
        dt = now()
    return (dt + relativedelta(day=31)).date()


def is_n_days_until_end_of_month(n_days: int) -> bool:
    """
    Return whether today + n days is the first of the next month.
    """
    n_days_from_now = today() + relativedelta(days=n_days)
    return n_days_from_now.day == 1


def is_date_within_date_range(
    date_in_question: datetime_.date, start: datetime_.date, end: datetime_.date
) -> bool:
    """
    Return whether a given date falls within the range of two other dates. This function assumes
    start < end.
    """
    return start <= date_in_question <= end


def is_in_the_past(dt: datetime_.datetime) -> bool:
    """
    Test whether a datetime is in the past.

    Note that we treat the current time as 'in the past' for the sake of this test. This means
    that any given datetime will always return True for either this function or
    `is_in_the_future`.
    """
    if not timezone.is_aware(dt):
        raise ValueError("Datetime must be timezone-aware")
    return dt <= now()


def is_in_the_future(dt: datetime_.datetime) -> bool:
    """
    Test whether a datetime is in the future.
    """
    if not timezone.is_aware(dt):
        raise ValueError("Datetime must be timezone-aware")
    return dt > now()


def is_future_date(_date: datetime_.date) -> bool:
    """
    Test whether a date is in the future.
    """
    return _date > today()


def is_within_the_last_year(date: datetime_.date) -> bool:
    """
    Test whether a date is one year or less before today's local date.
    """
    return today() - relativedelta(years=1) <= date <= today()


def is_within_the_last_week(date: datetime_.date) -> bool:
    """
    Test whether a date is within a week ago of today's dates inclusive of the earliest date.
    """
    return today() - relativedelta(days=7) <= date <= today()


def latest_date_for_day(
    start_date: datetime_.date, end_date: datetime_.date, day_of_month: int
) -> datetime_.date | None:
    """
    Given an integer day of a month, return the latest date with that day of the month,
    bounded by the supplied start_date and end_date. If no such date exists, return None.
    """
    if not (1 <= day_of_month <= 31):
        raise ValueError(f"{day_of_month} is not a valid day of the month.")

    date_range = ranges.FiniteRange(start_date, end_date, boundaries="[]")

    # Begin with a date in the same month as the end date.
    # (This will not necessarily be in range.)
    candidate_date = date_range.end + relativedelta(day=day_of_month)

    # Work our way backwards from the end date until we find a suitable date.
    while candidate_date >= date_range.start:
        if candidate_date.day == day_of_month and candidate_date in date_range:
            return candidate_date
        candidate_date -= relativedelta(months=1, day=day_of_month)
    return None


def next_date_with_day_of_month(
    date: datetime_.date, day_of_month: int
) -> datetime_.date:
    """
    Given a starting `date`, return the next date with the specified `day_of_month`.

    If the day of the month doesn't exist in the next month, return the nearest date that does.

    For example:

        next_date_with_day_of_month(date=date(2020, 1, 31), day_of_month=31)

        :returns: date(2020, 2, 29).

    :raises TypeError: if a datetime is used for `date`
    :raises ValueError: if the day of month is invalid

    Note: Python's datetime is a subclass of date, so the type checker will not prevent passing it
    to this function.
    """

    if isinstance(date, datetime_.datetime):
        raise TypeError("Must use a date, not a datetime.")

    if not (1 <= day_of_month <= 31):
        raise ValueError(f"{day_of_month} is not a valid day of the month.")

    next_date = date.replace(day=day_of_month)

    if next_date <= date:
        next_date += relativedelta(months=1, day=day_of_month)

    return next_date


def date_iterator(
    start_date: datetime_.date, end_date: datetime_.date
) -> Generator[datetime_.date, None, None]:
    """
    Iterate through dates between two dates
    """
    date = start_date

    while date < end_date:
        yield date
        date = date + datetime_.timedelta(days=1)


def system_timezone() -> tz.tzlocal:
    """
    Get the current system timezone.
    """
    return tz.tzlocal()


def is_dst(local_time: datetime_.datetime) -> bool:
    """
    Indicate whether the given time (and timezone is in daylight savings time (DST or not).

    Raises:
        ValueError if `local_time` is a naive datetime.
    """
    if timezone.is_naive(local_time):
        raise ValueError("Can't determine DST for a naive datetime")

    return bool(local_time.dst())


def is_localtime_midnight(
    dt: datetime_.datetime, tz: datetime_.tzinfo | None = None
) -> bool:
    """
    Return whether the supplied datetime is at midnight (in the site's local time zone).

    Note, the supplied datetime, which should be timezone aware, may be in any timezone,
    providing it corresponds to the moment of midnight in the site's local time zone.
    """
    return as_localtime(dt, tz).time() == datetime_.time(0)


def is_aligned_to_midnight(
    range: ranges.FiniteDatetimeRange, tz: datetime_.tzinfo | None = None
) -> bool:
    """
    Return whether this range is aligned to localtime midnight.
    """
    return all(
        [is_localtime_midnight(range.start, tz), is_localtime_midnight(range.end, tz)]
    )


def consolidate_into_intervals(
    dates: Sequence[datetime_.date],
) -> Sequence[Tuple[datetime_.date, datetime_.date]]:
    """
    Given a sequence of dates, return tuples of (inclusive) boundaries of the sub-sequences where
    the dates are consecutive.

    If a date does not form a continuous interval with any other dates - i.e. it is separated by at
    least a day before and after the others - it forms an interval with itself as both boundaries.

    For example, if passed:
    [
        <Date year=2020, month=1, day=1>,
        <Date year=2020, month=1, day=2>,
        <Date year=2020, month=1, day=4>,
        <Date year=2020, month=1, day=5>,
    ]
    it would return:
    [
        (<Date year=2020, month=1, day=1>, <Date year=2020, month=1, day=2>),
        (<Date year=2020, month=1, day=4>, <Date year=2020, month=1, day=5>),
    ]
    """
    if len(dates) < 1:
        raise ValueError("No dates provided.")

    dates = sorted(dates)
    intervals = []
    num_consecutive = 0

    interval_start, *remaining_dates = dates

    for i, date in enumerate(remaining_dates):
        if date == day_after(dates[i]):
            num_consecutive += 1
        else:
            intervals.append(
                (
                    interval_start,
                    interval_start + datetime_.timedelta(days=num_consecutive),
                )
            )
            interval_start = date
            num_consecutive = 0

    intervals.append(
        (interval_start, interval_start + datetime_.timedelta(days=num_consecutive))
    )

    return intervals


def translate_english_month_to_spanish(month: int) -> str:
    month_name = calendar.month_name[month]
    month_name_lookup = {
        "January": "enero",
        "February": "febrero",
        "March": "marzo",
        "April": "abril",
        "May": "mayo",
        "June": "junio",
        "July": "julio",
        "August": "agosto",
        "September": "septiembre",
        "October": "octubre",
        "November": "noviembre",
        "December": "deciembre",
    }
    return month_name_lookup[month_name]


def period_exceeds_one_year(
    start_at: datetime_.datetime, end_at: datetime_.datetime
) -> bool:
    """
    Returns true if the passed period exceeds one year.

    Edge cases such as leap years and daylight savings time are handled, where a simple approach
    using only relativedelta would not be sufficient.
    """

    # We take everything as localtime, and then remove the timezone information. This is to
    # avoid false results when the period starts or ends on a leap day, or when an uneven
    # number of DST changes happen in the year covered by the invoice.

    start_at = as_localtime(start_at)
    end_at = as_localtime(end_at)

    tz_unaware_start_at = start_at.replace(tzinfo=None)
    tz_unaware_end_at = end_at.replace(tzinfo=None)

    one_year_after_start_at = tz_unaware_start_at + relativedelta(years=1)

    if tz_unaware_start_at.month == 2 and tz_unaware_start_at.day == 29:
        # Leap year case, on 29th Feb.
        # One year after 29th Feb is 1st March (not 28th Feb).
        one_year_after_start_at = one_year_after_start_at.replace(month=3, day=1)

    return tz_unaware_end_at > one_year_after_start_at


def parse_date(value: str) -> datetime_.date:
    """
    Returns a datetime.date for a given ISO format date string.
    """
    return datetime_.date.fromisoformat(value)


def parse_dt(value: str, tz: datetime_.tzinfo | None = None) -> datetime_.datetime:
    """
    Returns a datetime.datetime for a given ISO format date/time string.

    The date/time must be naive - not include timezone information.
    """
    _datetime = datetime_.datetime.fromisoformat(value)
    return timezone.make_aware(_datetime, timezone=tz)


def strftime(
    dt: datetime_.datetime,
    fmt: str,
    tz: datetime_.tzinfo | None = None,
) -> str:
    """
    Format a tz aware datetime in localtime.

    Wrapper for `datetime.strftime`.
    """
    return as_localtime(dt, tz=tz).strftime(fmt)



from __future__ import annotations

import abc
import base64
import csv
import dataclasses
import datetime
import hashlib
import io
import os
import random
import re
import shutil
import urllib.parse
import uuid
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    AnyStr,
    BinaryIO,
    Iterable,
    Iterator,
    Protocol,
    TextIO,
    cast,
    runtime_checkable,
)

import boto3
import duckdb
import magic
import pandas as pd
from botocore import exceptions as botocore_exceptions
from botocore.response import StreamingBody
from django.conf import settings
from django.urls import reverse
from django.utils.module_loading import import_string
from pyarrow import parquet

from xocto import events, localtime

from . import files, s3_select


if TYPE_CHECKING:
    from _typeshed import WriteableBuffer
    from mypy_boto3_s3 import service_resource
    from mypy_boto3_s3.client import S3Client


# This URL is returned from `fetch_url()` in cases where a legitimate URL
# cannot be returned. For example, a legitimate URL cannot be returned
# for files stored in the `MemoryFileStore` during testing.
TEST_FETCH_URL = "http://www.example.com/file.txt"

TEMP_FOLDER = "/tmp"  # nosec

# The maximum length allowed for an S3 key
MAX_KEY_LENGTH = 1024

# S3 select can have a max size of 1 MB.
MAX_S3_SELECT_SIZE_RANGE = 1_048_576

# Regex for s3 URLs in the "virtual hosted" format
S3_VIRTUAL_HOSTED_URL_RE = re.compile(
    r"https://(?P<bucket>.+).s3.((?P<region>.*).)?amazonaws.com/(?P<key>.+)"
)


def _should_raise_error_on_existing_files() -> bool:
    """
    Check if we should error when a file is being overwritten without explicit overwrite=True.
    """
    return False


def _log_existing_file_returned(filename: str) -> None:
    events.publish("storage.file.existing-returned", params={"filename": filename})


class FileExists(Exception):
    pass


class KeyDoesNotExist(Exception):
    pass


class BucketNotVersioned(Exception):
    pass


class S3SelectUnexpectedResponse(Exception):
    pass


@dataclasses.dataclass(frozen=True)
class PreSignedPost:
    """
    Pre-signed post url.

    For more details see:
    https://boto3.amazonaws.com/v1/documentation/api/latest/guide/s3-presigned-urls.html#generating-a-presigned-url-to-upload-a-file
    """

    url: str
    key: str
    fields: dict[str, Any]


@dataclasses.dataclass(frozen=True)
class S3Object:
    bucket_name: str
    key: str
    version_id: str | None = None


@runtime_checkable
class Clearable(Protocol):
    def clear(self) -> None: ...


class ReadableBinaryFile(Protocol):
    def read(self, size: int = ...) -> bytes: ...


class StreamingBodyIOAdapter(io.RawIOBase):
    """
    Wrapper to adapt a boto3 S3 object to a standard Python "file-like" object

    Boto3 returns S3 files as instances as of botocore.response.StreamingBody. These have a `read`
    method but are otherwise not quite "file-like" enough to be drop in equivalents of anything in
    the normal Python IO hierarchy.

    This class is an subclass of io.RawIOBase which works by wrapping a boto StreamingBody. The
    storage module uses it so that it can return normal looking Python file objects and hide
    whether the backing store is S3 or not.
    """

    def __init__(self, streaming_body: StreamingBody) -> None:
        self.streaming_body = streaming_body

    def readable(self) -> bool:
        return True

    def readinto(self, buffer: WriteableBuffer) -> int:
        """
        Read bytes into a pre-allocated, writable buffer-like object

        Mutates `buffer` and returns the number of bytes read.
        """
        # buffer has to be typed as `buffer: WriteableBuffer` because that's how the superclass is
        # typed. However, WriteableBuffer is a Union of several Python buffer-like objects, not all
        # of which support __len__ or __set_item__ in the way needed by our implementation of this
        # method. This assert exists to make mypy pass but isn't totally satisfying because in
        # theory there might be a way that this method is called with one of the weirder buffer
        # classes in the Union, like pickle.PickleBuffer.
        # See https://github.com/python/typeshed/blob/master/stdlib/_typeshed/__init__.pyi for the
        # definition of WriteableBuffer.
        assert isinstance(buffer, (bytearray, memoryview))

        requested_size = len(buffer)
        data = self.streaming_body.read(requested_size)
        data_size = len(data)
        buffer[:data_size] = data
        return data_size

    def close(self) -> None:
        self.streaming_body.close()
        super().close()


class BaseS3FileStore(abc.ABC):
    # Define the interface for subclasses

    def __init__(
        self,
        bucket_name: str,
        use_date_in_key_path: bool = True,
        set_acl_bucket_owner: bool = False,
    ) -> None:
        self.bucket_name = bucket_name
        self.date_in_key_path = use_date_in_key_path
        self.set_acl_bucket_owner = set_acl_bucket_owner

    @abc.abstractmethod
    def store_file(
        self,
        namespace: str,
        filename: str,
        contents: AnyStr | ReadableBinaryFile,
        content_type: str = "",
        overwrite: bool = False,
        metadata: dict[str, str] | None = None,
    ) -> tuple[str, str]:
        """
        Store a file in S3 given its filename and contents. Contents should be UTF-8 encoded.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def store_versioned_file(
        self,
        key_path: str,
        contents: AnyStr | io.BytesIO,
        content_type: str = "",
    ) -> tuple[str, str, str]:
        """
        Store a file in S3 given its filename and contents. Contents should be UTF-8 encoded.

        The bucket must have versioning enabled.
        If the key is not known, `make_key_path` should be called to generate it.

        :raises BucketNotVersioned: if the bucket does not have versioning enabled.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def store_filepath(
        self,
        namespace: str,
        filepath: str,
        overwrite: bool = False,
        dest_filepath: str = "",
    ) -> tuple[str, str]:
        raise NotImplementedError()

    def make_key_path(self, *, namespace: str = "", filepath: str) -> str:
        """
        Return the full file (key) path given a namespace and filepath.

        This is normally just `{namespace}/{filepath}` except when `self.use_date_in_key_path` is
        true, in which case today's date is inserted in between:

            `{namespace}/2021/07/20/{filepath}`

        Use this to determine the key path that would be returned by the `store_file*` functions
        for the same namespace and filepath.
        """
        if namespace:
            # Remove trailing slash to avoid creating a "directory" named "/" in the path.
            namespace = namespace.rstrip("/")

        if self.date_in_key_path:
            today = localtime.today()
            parts = [v for v in [namespace, today.strftime("%Y/%m/%d"), filepath] if v]
        else:
            parts = [v for v in [namespace, filepath] if v]

        key_path = os.path.join(*parts)
        if len(key_path) > MAX_KEY_LENGTH:
            raise RuntimeError(
                f"Generated `key_path` must not exceed {MAX_KEY_LENGTH} characters in length"
            )

        return key_path

    @abc.abstractmethod
    def get_key(self, key_path: str, version_id: str | None = None) -> S3Object:
        raise NotImplementedError()

    def get_key_or_store_file(
        self,
        *,
        namespace: str = "",
        filepath: str,
        contents: bytes,
        content_type: str = "",
    ) -> tuple[tuple[str, str], bool]:
        """
        Return the full key for `namespace` and `filepath`, writing `contents` if there's nothing
        already at that key path.

        Warning: this is open to race conditions for certain storage backends (S3) because there's
        no way to transactionally create an object at a key. E.g.:

            * Process 1 does get_key_or_store_file
            * Process 1 sees there's no file
            * Before process 1 can create the file at the key
            * Process 2 does get_key_or_store_file
            * Process 2 also sees there's no file
            * Whichever process finishes the writing to the key last wins
        """
        key_path = self.make_key_path(namespace=namespace, filepath=filepath)

        if self.exists(key_path):
            return (self.bucket_name, key_path), False

        self.store_file(
            namespace=namespace,
            filename=filepath,
            contents=contents,
            content_type=content_type,
        )
        return (self.bucket_name, key_path), True

    @abc.abstractmethod
    def get_file_type(self, key_path: str) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def fetch_file(
        self, key_path: str, version_id: str | None = None, range: str | None = None
    ) -> StreamingBody:
        raise NotImplementedError()

    @abc.abstractmethod
    def fetch_file_contents(
        self, key_path: str, version_id: str | None = None
    ) -> bytes:
        raise NotImplementedError()

    def fetch_text_file(
        self, key_path: str, encoding: str | None = None, errors: str | None = None
    ) -> TextIO:
        """
        Return a file from storage as a TextIO "file-like" object.
        """
        streaming_body = self.fetch_file(key_path)
        raw_io = StreamingBodyIOAdapter(streaming_body)
        buffered_io = io.BufferedReader(raw_io)
        return io.TextIOWrapper(buffered_io, encoding=encoding, errors=errors)

    @abc.abstractmethod
    def fetch_url(
        self,
        key_path: str,
        expires_in: int = 60,
        response_headers: dict[str, str] | None = None,
        version_id: str | None = None,
    ) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    def generate_presigned_post(
        self, *, key_path: str, expires_in: int = 60
    ) -> PreSignedPost:
        raise NotImplementedError

    @abc.abstractmethod
    def exists(self, key_path: str, as_file: bool = False) -> bool:
        raise NotImplementedError()

    @abc.abstractmethod
    def list_s3_keys(self, namespace: str = "") -> Iterable[S3Object]:
        raise NotImplementedError()

    def list_s3_keys_page(  # nosec B107
        self,
        namespace: str = "",
        *,
        next_token: str = "",
        max_keys: int = 100,
    ) -> tuple[Iterable[S3Object], str]:
        """Pass-thru to list_s3_keys"""
        return self.list_s3_keys(namespace), ""

    @abc.abstractmethod
    def list_files(self, namespace: str = "") -> Iterable[str]:
        raise NotImplementedError()

    def download_file(self, key_path: str) -> BinaryIO:
        filepath = self._build_download_filepath(key_path)
        _create_parent_directories(filepath)
        with open(filepath, "wb") as f:
            f.write(self.fetch_file_contents(key_path))
        return f

    def download_to_file(self, key_path: str, file_path: str) -> None:
        with open(file_path, "wb") as f:
            f.write(self.fetch_file_contents(key_path))

    @abc.abstractmethod
    def get_size_in_bytes(self, *, s3_object: S3Object) -> int:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_last_modified(self, key_path: str) -> datetime.datetime:
        raise NotImplementedError()

    @abc.abstractmethod
    def copy(self, *, s3_object: S3Object, destination: str) -> S3Object:
        raise NotImplementedError()

    @abc.abstractmethod
    def rename(self, *, s3_object: S3Object, destination: str) -> S3Object:
        """
        Rename an S3 object.

        The behaviour in a versioned bucket is undefined.
        It should be defined once a clear use case has been identified.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self, *, s3_object: S3Object) -> None:
        """
        Delete an object from S3.

        If the bucket is versioned, this will only delete the specified object version.
        """
        raise NotImplementedError()

    def _build_download_filepath(self, key_path: str) -> str:
        """
        Assemble the full filename to create when downloading a file.

        Args:
            key_path (string): the S3 key of the file that will be downloaded.
        """
        return os.path.join(TEMP_FOLDER, self.bucket_name, key_path)


class S3FileStore(BaseS3FileStore):
    ACL_BUCKET_OWNER_FULL_CONTROL = "bucket-owner-full-control"

    def __init__(
        self,
        bucket_name: str,
        use_date_in_key_path: bool = True,
        set_acl_bucket_owner: bool = False,
    ) -> None:
        if not (3 <= len(bucket_name) <= 63):
            raise ValueError(
                f"`bucket_name` must be between 3 and 63 characters in length: {bucket_name}"
            )
        super().__init__(
            bucket_name,
            use_date_in_key_path=use_date_in_key_path,
            set_acl_bucket_owner=set_acl_bucket_owner,
        )

    def __str__(self) -> str:
        return f"S3 FileStore for bucket {self.bucket_name}"

    def store_file(
        self,
        namespace: str,
        filename: str,
        contents: AnyStr | ReadableBinaryFile,
        content_type: str = "",
        overwrite: bool = False,
        metadata: dict[str, str] | None = None,
    ) -> tuple[str, str]:
        """
        Store a file in S3 given its filename and contents. Contents should be UTF-8 encoded.

        :raises FileExists: if a file with the given name already exists in the bucket and
        raising an exception is enabled via an env var.
        """
        key_path = self.make_key_path(namespace=namespace, filepath=filename)

        if not overwrite:
            try:
                existing_boto_object = self._get_boto_object_for_key(key=key_path)
            except KeyDoesNotExist:
                pass
            else:
                if _should_raise_error_on_existing_files():
                    raise FileExists(
                        "A file with this name already exists. Pass overwrite=True if you're sure "
                        "it's safe to overwrite the contents of the existing file."
                    )
                _log_existing_file_returned(key_path)
                return self.bucket_name, existing_boto_object.key

        readable = _to_stream(contents=contents)

        # `boto_client.upload_fileobj` is type annotated with `Fileobj: BinaryIO`. However, in
        # practice the only file-like method it needs is `read(size=...)`. This cast allows us to
        # use `upload_fileobj` with any Fileobj that implements the `ReadableBinaryFile` protocol
        # but not the whole of BinaryIO. This includes, importantly, Django's `UploadedFile`.
        file_obj = cast(BinaryIO, readable)

        extra_args: dict[str, Any] = {}
        if content_type:
            extra_args["ContentType"] = content_type
        if policy := self._get_policy():
            extra_args["ACL"] = policy
        if metadata:
            extra_args["Metadata"] = metadata

        boto_client = self._get_boto_client()
        boto_client.upload_fileobj(
            Fileobj=file_obj,
            Bucket=self.bucket_name,
            Key=key_path,
            ExtraArgs=extra_args,
        )

        return self.bucket_name, key_path

    def store_versioned_file(
        self,
        key_path: str,
        contents: AnyStr | io.BytesIO,
        content_type: str = "",
    ) -> tuple[str, str, str]:
        if not self._bucket_is_versioned():
            raise BucketNotVersioned()

        file_obj = _to_stream(contents=contents)

        extra_args: dict[str, str] = {}
        if content_type:
            extra_args["ContentType"] = content_type
        if policy := self._get_policy():
            extra_args["ACL"] = policy

        boto_client = self._get_boto_client()
        boto_response = boto_client.put_object(
            Body=file_obj,
            Bucket=self.bucket_name,
            Key=key_path,
            **extra_args,  # type: ignore[arg-type]
        )
        version_id = boto_response["VersionId"]

        return self.bucket_name, key_path, version_id

    def store_filepath(
        self,
        namespace: str,
        filepath: str,
        overwrite: bool = False,
        dest_filepath: str = "",
    ) -> tuple[str, str]:
        """
        Store a file in S3 given its local filepath.
        """
        if not dest_filepath:
            dest_filepath = os.path.basename(filepath)

        key_path = self.make_key_path(namespace=namespace, filepath=dest_filepath)

        if not overwrite:
            try:
                existing_boto_object = self._get_boto_object_for_key(key=key_path)
            except KeyDoesNotExist:
                pass
            else:
                if _should_raise_error_on_existing_files():
                    raise FileExists(
                        "A file with this name already exists. Pass overwrite=True if you're sure "
                        "it's safe to overwrite the contents of the existing file."
                    )
                _log_existing_file_returned(key_path)
                return self.bucket_name, existing_boto_object.key

        extra_args = {}
        if policy := self._get_policy():
            extra_args["ACL"] = policy

        boto_client = self._get_boto_client()
        boto_client.upload_file(
            Filename=filepath,
            Bucket=self.bucket_name,
            Key=key_path,
            ExtraArgs=extra_args,
        )

        return self.bucket_name, key_path

    def get_key(self, key_path: str, version_id: str | None = None) -> S3Object:
        return S3Object(
            bucket_name=self.bucket_name, key=key_path, version_id=version_id
        )

    def get_file_type(self, key_path: str) -> str:
        return self._get_boto_object_for_key(key=key_path).content_type

    def fetch_file(
        self, key_path: str, version_id: str | None = None, range: str | None = None
    ) -> StreamingBody:
        boto_object = self._get_boto_object_for_key(key=key_path, version_id=version_id)
        if range:
            return boto_object.get(Range=range)["Body"]
        return boto_object.get()["Body"]

    def fetch_file_contents(
        self, key_path: str, version_id: str | None = None
    ) -> bytes:
        return self.fetch_file(key_path, version_id).read()

    def fetch_file_contents_using_s3_select(
        self,
        key_path: str,
        raw_sql: str,
        input_serializer: s3_select.CSVInputSerializer
        | s3_select.ParquetInputSerializer,
        output_serializer: s3_select.CSVOutputSerializer
        | s3_select.JSONOutputSerializer,
        compression_type: s3_select.CompressionType | None = None,
        scan_range: s3_select.ScanRange | None = None,
        chunk_size: int | None = None,
    ) -> Iterator[str]:
        """
        Reads a CSV or Parquet file from S3 using the given SQL statement.
        Reference: https://dev.to/idrisrampurawala/efficiently-streaming-a-large-aws-s3-file-via-s3-select-4on
        Does not support columnar compression for parquet files at present.
        """

        boto_client = self._get_boto_client()

        if isinstance(input_serializer, s3_select.CSVInputSerializer):
            serialization = s3_select.get_serializers_for_csv_file(
                input_serializer=input_serializer,
                compression_type=compression_type
                if compression_type is not None
                else s3_select.CompressionType.NONE,
                output_serializer=output_serializer,
                scan_range=scan_range,
            )
        elif isinstance(input_serializer, s3_select.ParquetInputSerializer):
            if scan_range is not None:
                raise ValueError(
                    "The scan_range parameter is not supported for parquet files"
                )
            serialization = s3_select.get_serializers_for_parquet_file(
                output_serializer=output_serializer
            )
        else:
            raise ValueError(
                "input_serializer must be either CSVInputSerializer or ParquetInputSerializer"
            )

        select_object_content_parameters = dict(
            Bucket=self.bucket_name,
            Key=key_path,
            ExpressionType="SQL",
            Expression=raw_sql,
            InputSerialization=serialization["input_serialization"],
            OutputSerialization=serialization["output_serialization"],
        )

        if scan_range:
            yield from self._select_object_content_using_scan_range(
                boto_client=boto_client,
                select_object_content_parameters=select_object_content_parameters,
                key_path=key_path,
                scan_range=scan_range,
                chunk_size=chunk_size,
            )
        else:
            yield from self._select_object_content(
                boto_client=boto_client,
                select_object_content_parameters=select_object_content_parameters,
            )

    def fetch_url(
        self,
        key_path: str,
        expires_in: int = 60,
        response_headers: dict[str, str] | None = None,
        version_id: str | None = None,
    ) -> str:
        """
        Return a presigned URL that grants public access to an object. A presigned URL remains
        valid for a limited period of time which is specified with `expires_in` in seconds.
        Additional response headers can be provided with `response_headers`, using keys such as
        `"ResponseContentType"` or `"ResponseContentDisposition"`.
        """
        params = response_headers or {}
        # Ensure these values take precedence if the keys were ever, incorrectly, already in
        # the `response_headers` dictionary.
        params.update({"Bucket": self.bucket_name, "Key": key_path})
        if version_id:
            params.update({"VersionId": version_id})
        boto_client = self._get_boto_client()
        return boto_client.generate_presigned_url(
            "get_object", Params=params, ExpiresIn=expires_in
        )

    def generate_presigned_post(
        self, *, key_path: str, expires_in: int = 60
    ) -> PreSignedPost:
        boto_client = self._get_boto_client()
        presigned_post = boto_client.generate_presigned_post(
            Bucket=self.bucket_name, Key=key_path, ExpiresIn=expires_in
        )
        return PreSignedPost(
            url=presigned_post["url"],
            fields=presigned_post["fields"],
            key=presigned_post["fields"]["key"],
        )

    def exists(self, key_path: str, as_file: bool = False) -> bool:
        try:
            boto_object = self._get_boto_object_for_key(key=key_path)
        except KeyDoesNotExist:
            return False
        return not (as_file and boto_object.content_length == 0)

    def list_s3_keys(self, namespace: str = "") -> Iterable[S3Object]:
        boto_bucket = self._get_boto_bucket()
        for object_summary in boto_bucket.objects.filter(Prefix=namespace):
            yield S3Object(bucket_name=self.bucket_name, key=object_summary.key)

    def list_s3_keys_page(  # nosec B107
        self,
        namespace: str = "",
        *,
        next_token: str = "",
        max_keys: int = 100,
    ) -> tuple[Iterable[S3Object], str]:
        """
        Lists subset of files in the S3 bucket, optionally limited to only those within a given
        namespace. S3 keys ending with a forward slash are excluded, on the assumption that they
        are folders, not files.

        :param namespace: Limits keys to only those with this prefix
        :param next_token: next_token supplied from the last returned result
        :param max_keys: the number of keys to request from the bucket (<1000)
        :return: Tuple of generator of key names, and next token for next page
        """
        client = self._get_boto_client()
        # Not using ContinuationToken as ContinuationToken is tied to a specific client
        response = client.list_objects_v2(
            Bucket=self.bucket_name,
            Prefix=namespace,
            MaxKeys=max_keys,
            StartAfter=next_token,
        )
        keys = [item["Key"] for item in response.get("Contents") or []]
        next_token = keys[-1] if response["IsTruncated"] else ""
        keys = [key for key in keys if not key.endswith("/")]
        objects = [S3Object(bucket_name=self.bucket_name, key=key) for key in keys]
        return objects, next_token

    def list_files(self, namespace: str = "") -> Iterable[str]:
        """
        Lists all files in the S3 bucket, optionally limited to only those within a given
        namespace. S3 keys ending with a forward slash are excluded, on the assumption that they
        are folders, not files.

        :param namespace: Limits keys to only those with this prefix
        :return: Generator of key names
        """
        boto_bucket = self._get_boto_bucket()
        for object_summary in boto_bucket.objects.filter(Prefix=namespace):
            if not object_summary.key.endswith("/"):  # filter out folders
                yield object_summary.key

    def download_file(self, key_path: str) -> BinaryIO:
        """
        A more efficient version of the superclass's method.
        """
        boto_object = self._get_boto_object_for_key(key=key_path)
        filepath = self._build_download_filepath(key_path)
        _create_parent_directories(filepath)
        with open(filepath, "wb") as f:
            boto_object.download_fileobj(Fileobj=f)
        return f

    def download_to_file(self, key_path: str, file_path: str) -> None:
        boto_object = self._get_boto_object_for_key(key=key_path)
        boto_object.download_file(Filename=file_path)

    def get_size_in_bytes(self, *, s3_object: S3Object) -> int:
        boto_object = self._get_boto_object(s3_object=s3_object)
        return boto_object.content_length

    def get_last_modified(self, key_path: str) -> datetime.datetime:
        boto_object = self._get_boto_object_for_key(key=key_path)
        return boto_object.last_modified

    def copy(self, *, s3_object: S3Object, destination: str) -> S3Object:
        extra_args = {}
        if policy := self._get_policy():
            extra_args["ACL"] = policy
        dest_boto_object = self._get_boto_bucket().Object(destination)
        dest_boto_object.copy(
            CopySource={"Bucket": s3_object.bucket_name, "Key": s3_object.key},
            ExtraArgs=extra_args,
        )
        return S3Object(bucket_name=self.bucket_name, key=destination)

    def rename(self, *, s3_object: S3Object, destination: str) -> S3Object:
        src_boto_object = self._get_boto_object(s3_object=s3_object)
        extra_args = {}
        if policy := self._get_policy():
            extra_args["ACL"] = policy
        dest_boto_object = self._get_boto_bucket().Object(destination)
        dest_boto_object.copy(
            CopySource={"Bucket": s3_object.bucket_name, "Key": s3_object.key},
            ExtraArgs=extra_args,
        )
        if src_boto_object.key != dest_boto_object.key:
            # Only delete the old file if the source and destination are different. Otherwise
            # renaming a file to its own path would delete it.
            src_boto_object.delete()
        return S3Object(bucket_name=self.bucket_name, key=destination)

    def delete(self, *, s3_object: S3Object) -> None:
        boto_object = self._get_boto_object(s3_object=s3_object)
        boto_object.delete()

    # Private

    def _get_policy(self) -> str | None:
        if self.set_acl_bucket_owner:
            # If the storage class is configured to, we will set the ACL of keys that we
            # set content on to have the policy "bucket-owner-full-control".
            # This is useful when writing to a bucket where the owner is an external AWS account.
            # NB: You will need PutObjectAcl permissions on the bucket in order to set the ACL.
            return self.ACL_BUCKET_OWNER_FULL_CONTROL
        return None

    def _get_boto_client(self) -> S3Client:
        return boto3.client(
            "s3",
            region_name=settings.AWS_REGION,
            endpoint_url=settings.AWS_S3_ENDPOINT_URL,
        )

    def _get_boto_bucket(self) -> service_resource.Bucket:
        boto_resource = boto3.resource("s3", region_name=settings.AWS_REGION)
        return boto_resource.Bucket(self.bucket_name)

    def _bucket_is_versioned(self) -> bool:
        boto_client = self._get_boto_client()
        versioning_info = boto_client.get_bucket_versioning(Bucket=self.bucket_name)
        return versioning_info["Status"] == "Enabled"

    def _get_boto_object(self, *, s3_object: S3Object) -> service_resource.Object:
        assert s3_object.bucket_name == self.bucket_name, (
            f"Expected an S3Object from the '{self.bucket_name}' store, "
            f"but it's from the '{s3_object.bucket_name}' store"
        )
        boto_object = self._get_boto_bucket().Object(key=s3_object.key)
        if s3_object.version_id:
            if not self._bucket_is_versioned():
                raise BucketNotVersioned()
            boto_object.version_id = s3_object.version_id

        try:
            # Calls `S3.Client.head_object()` to fetch the object's attributes; e.g. its size.
            boto_object.load()
        except botocore_exceptions.ClientError as error:
            if error.response.get("Error", {}).get("Code", "") == "404":
                raise KeyDoesNotExist("Key with path %s was not found" % s3_object.key)
            raise
        return boto_object

    def _get_boto_object_for_key(
        self, *, key: str, version_id: str | None = None
    ) -> service_resource.Object:
        # Note: It looks like it can be DRYed up with a call to self.get_key().
        # *This is not safe.* It cannot be DRYed like this.
        # It'd cause the S3SubdirectoryFileStore to add the subdir to the path, which is
        # not safe since the key may have come from somewhere that already includes this.
        return self._get_boto_object(
            s3_object=S3Object(
                bucket_name=self.bucket_name, key=key, version_id=version_id
            )
        )

    def _select_object_content(
        self,
        *,
        boto_client: S3Client,
        select_object_content_parameters: dict[str, Any],
    ) -> Iterator[str]:
        # Error codes reference: https://docs.aws.amazon.com/AmazonS3/latest/API/ErrorResponses.html#SelectObjectContentErrorCodeList
        invalid_response_statuses = [400, 401, 403, 500]

        try:
            response = boto_client.select_object_content(
                **select_object_content_parameters,
            )
        except botocore_exceptions.ClientError as error:
            if (
                error.response.get("Error", {}).get("HTTPStatusCode", None)
                in invalid_response_statuses
            ):
                raise S3SelectUnexpectedResponse(
                    "Received invalid response from S3 Select"
                )
            raise

        if response["ResponseMetadata"]["HTTPStatusCode"] in invalid_response_statuses:
            raise S3SelectUnexpectedResponse("Received invalid response from S3 Select")

        for event_stream in response["Payload"]:
            if records := event_stream.get("Records"):
                yield records["Payload"].decode("utf-8")

    def _select_object_content_using_scan_range(
        self,
        *,
        boto_client: S3Client,
        select_object_content_parameters: dict[str, Any],
        key_path: str,
        scan_range: s3_select.ScanRange,
        chunk_size: int | None = None,
    ) -> Iterator[str]:
        """
        Performs SQL queries on S3 objects (CSV/JSON) using the given offsets (Scan Range)
        """
        file_size = self.get_size_in_bytes(
            s3_object=S3Object(bucket_name=self.bucket_name, key=key_path)
        )

        start_range = scan_range.Start if scan_range.Start else 0

        if chunk_size:
            chunk_size = min(chunk_size, MAX_S3_SELECT_SIZE_RANGE)
            end_range = scan_range.End if scan_range.End else min(chunk_size, file_size)

            while start_range < file_size:
                yield from self._select_object_content(
                    boto_client=boto_client,
                    select_object_content_parameters=dict(
                        **select_object_content_parameters,
                        ScanRange=dataclasses.asdict(
                            s3_select.ScanRange(Start=start_range, End=end_range)
                        ),
                    ),
                )
                start_range = end_range
                end_range = end_range + min(chunk_size, file_size - end_range)
        else:
            end_range = scan_range.End if scan_range.End else file_size

            if (end_range - start_range) > MAX_S3_SELECT_SIZE_RANGE:
                raise ValueError(
                    f"The difference between the start range and end range should be less than 1 MB ({MAX_S3_SELECT_SIZE_RANGE} bytes)."
                )

            yield from self._select_object_content(
                boto_client=boto_client,
                select_object_content_parameters=dict(
                    **select_object_content_parameters,
                    ScanRange=dataclasses.asdict(
                        s3_select.ScanRange(Start=start_range, End=end_range)
                    ),
                ),
            )


class S3SubdirectoryFileStore(S3FileStore):
    """
    A S3FileStore which can expose just a given subdirectory rather than a whole bucket.
    """

    def __init__(self, uri: str) -> None:
        parsed_url = urllib.parse.urlparse(uri)
        if parsed_url.scheme != "s3":
            raise ValueError(f"Expected URL starting 's3://'. Got {uri!r}")
        if not parsed_url.netloc:
            raise ValueError(f"Expected S3 URL including a bucket name. Got {uri!r}")

        set_acl_bucket_owner = "set_acl_bucket_owner" in parsed_url.query
        use_date_in_key_path = "use_date_in_key_path" in parsed_url.query

        self.bucket_name = parsed_url.netloc
        self.path = parsed_url.path.strip("/")
        super().__init__(
            self.bucket_name,
            use_date_in_key_path=use_date_in_key_path,
            set_acl_bucket_owner=set_acl_bucket_owner,
        )

    def make_key_path(self, *, namespace: str = "", filepath: str) -> str:
        if self.path:
            namespace = os.path.join(self.path, namespace) if namespace else self.path
        return super().make_key_path(namespace=namespace, filepath=filepath)

    def get_key(self, key_path: str, version_id: str | None = None) -> S3Object:
        if self.path:
            key_path = os.path.join(self.path, key_path)
        return super().get_key(key_path, version_id)

    def fetch_url(
        self,
        key_path: str,
        expires_in: int = 60,
        response_headers: dict[str, str] | None = None,
        version_id: str | None = None,
    ) -> str:
        if self.path:
            key_path = os.path.join(self.path, key_path)
        return super().fetch_url(
            key_path=key_path,
            expires_in=expires_in,
            response_headers=response_headers,
            version_id=version_id,
        )

    def list_s3_keys(self, namespace: str = "") -> Iterable[S3Object]:
        if self.path:
            namespace = os.path.join(self.path, namespace)
        return super().list_s3_keys(namespace)

    def list_s3_keys_page(
        self,
        namespace: str = "",
        *,
        next_token: str = "",
        max_keys: int = 100,
    ) -> tuple[Iterable[S3Object], str]:
        if self.path:
            namespace = os.path.join(self.path, namespace)
        objects, next_token = super().list_s3_keys_page(
            namespace, next_token=next_token, max_keys=max_keys
        )
        if self.path:
            objects = [
                dataclasses.replace(object, key=object.key[len(self.path) + 1 :])
                for object in objects
            ]
        return objects, next_token

    def list_files(self, namespace: str = "") -> Iterable[str]:
        path_trim = 0
        if self.path:
            path_trim = len(self.path) + 1
            namespace = os.path.join(self.path, namespace)
        full_paths = super().list_files(namespace)
        yield from (path[path_trim:] for path in full_paths)

    def copy(self, *, s3_object: S3Object, destination: str) -> S3Object:
        if self.path:
            destination = os.path.join(self.path, destination)
        return super().copy(s3_object=s3_object, destination=destination)

    def rename(self, *, s3_object: S3Object, destination: str) -> S3Object:
        if self.path:
            destination = os.path.join(self.path, destination)
        return super().rename(s3_object=s3_object, destination=destination)

    # Private

    def _get_boto_object_for_key(
        self, *, key: str, version_id: str | None = None
    ) -> service_resource.Object:
        if self.path and not key.startswith(self.path):
            key = os.path.join(self.path, key)
        return super()._get_boto_object_for_key(key=key, version_id=version_id)


class LocalStorageBucket:
    """
    S3 bucket for local development.
    """

    document_url_base = "/static/support/sample-files/sample-documents"
    document_filenames = ["sample.pdf"]


class LocalFileStore(BaseS3FileStore):
    """
    For local development.
    """

    def __init__(
        self,
        bucket_name: str,
        storage_root: str = "",
        use_date_in_key_path: bool = True,
        set_acl_bucket_owner: bool = False,
    ) -> None:
        # This is taken from https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucketnamingrules.html
        if not (3 <= len(bucket_name) <= 63):
            raise ValueError(
                f"`bucket_name` must be between 3 and 63 characters in length: {bucket_name}"
            )

        # N.B. Overriding the storage root will mean that fetch_url won't work.
        self.storage_root = storage_root or settings.MEDIA_ROOT or "/tmp"  # nosec B108
        super().__init__(
            bucket_name,
            use_date_in_key_path=use_date_in_key_path,
            set_acl_bucket_owner=set_acl_bucket_owner,
        )

    def __str__(self) -> str:
        return f"Local FileStore for bucket {self.bucket_name} with root {self.storage_root}"

    def store_file(
        self,
        namespace: str,
        filename: str,
        contents: AnyStr | ReadableBinaryFile,
        content_type: str = "",
        overwrite: bool = False,
        metadata: dict[str, str] | None = None,
    ) -> tuple[str, str]:
        filepath = self._filepath(namespace, filename)
        if os.path.exists(filepath) and not overwrite:
            raise FileExists(
                f"A file with this name, {filename} already exists. Pass overwrite=True if you're sure "
                "it's safe to overwrite the contents of the existing file."
            )

        with open(filepath, "wb") as f:
            f.write(_to_bytes(contents=contents))

        return self.bucket_name, filepath

    def store_versioned_file(
        self,
        key_path: str,
        contents: AnyStr | io.BytesIO,
        content_type: str = "",
    ) -> tuple[str, str, str]:
        version = str(uuid.uuid4())
        key_components = os.path.split(key_path)
        versioned_key = os.path.join(key_components[0], version, key_components[1])
        versioned_filepath = self._filepath_for_key_path(versioned_key)

        filepath = self._filepath_for_key_path(key_path)

        with open(versioned_filepath, "wb") as f:
            f.write(_to_bytes(contents=contents))
        with open(filepath, "wb") as f:
            f.write(_to_bytes(contents=contents))

        return self.bucket_name, filepath, version

    def _build_download_filepath(self, key_path: str) -> str:
        # When uploading a file locally, the filepath we upload to is identical to the location
        # we then attempt to download to. To avoid overwriting the file, we "download" to a
        # separate "downloaded" bucket
        key_path = key_path.replace(self.bucket_name, f"{self.bucket_name}-downloaded")
        return os.path.join(self.storage_root, key_path)

    def store_filepath(
        self,
        namespace: str,
        filepath: str,
        overwrite: bool = False,
        dest_filepath: str = "",
    ) -> tuple[str, str]:
        if not dest_filepath:
            dest_filepath = os.path.basename(filepath)

        store_filepath = self._filepath(namespace, dest_filepath)
        if os.path.exists(store_filepath) and not overwrite:
            raise FileExists(
                "A file with this name already exists. Pass overwrite=True if you're sure "
                "it's safe to overwrite the contents of the existing file."
            )

        shutil.copyfile(filepath, store_filepath)
        return self.bucket_name, store_filepath

    def get_key(self, key_path: str, version_id: str | None = None) -> S3Object:
        return S3Object(
            bucket_name=self.bucket_name, key=key_path, version_id=version_id
        )

    def get_file_type(self, key_path: str) -> str:
        mime = magic.Magic(mime=True)
        return mime.from_buffer(self.fetch_file_contents(key_path))

    def fetch_file(
        self, key_path: str, version_id: str | None = None, range: str | None = None
    ) -> StreamingBody:
        # If given an absolute path, use that, otherwise combine with the root and bucket. When
        # fetching an object from a path, we don't need to interpolate the date or namespace.
        if version_id:
            path_components = os.path.split(key_path)
            file_path = os.path.join(
                self.storage_root,
                self.bucket_name,
                path_components[0],
                version_id,
                path_components[1],
            )
        else:
            file_path = os.path.join(self.storage_root, self.bucket_name, key_path)
        if not os.path.exists(file_path):
            raise KeyDoesNotExist(f"Key {key_path} was not found at {file_path}")
        with open(file_path, "rb") as f:
            raw_stream = io.BytesIO(f.read())
        return StreamingBody(
            raw_stream=raw_stream, content_length=files.size(raw_stream)
        )

    def fetch_file_contents(
        self, key_path: str, version_id: str | None = None
    ) -> bytes:
        return self.fetch_file(key_path, version_id).read()

    def fetch_file_contents_using_s3_select(
        self,
        key_path: str,
        raw_sql: str,
        input_serializer: s3_select.CSVInputSerializer
        | s3_select.ParquetInputSerializer,
        output_serializer: s3_select.CSVOutputSerializer
        | s3_select.JSONOutputSerializer,
        compression_type: s3_select.CompressionType | None = None,
        scan_range: s3_select.ScanRange | None = None,
        chunk_size: int | None = None,
    ) -> Iterator[str]:
        """
        Localdev version of S3FileStore `fetch_file_contents_using_s3_select`. Does not support scan ranges.
        Converts the file to a pandas dataframe and then runs the SQL query on it
        by creating an in memory DB with the table which a query can be run on. As the file is loaded into memory,
        this is not recommended for use with very large files.

        While this should return the same results as the S3FileStore version,
        there may be small discrepancies in the results being returned.
        """

        if scan_range is not None:
            raise NotImplementedError("Scan ranges are not supported for localdev")

        input_file_path = self._filepath_for_key_path(key_path=key_path)

        if isinstance(input_serializer, s3_select.CSVInputSerializer):
            # https://docs.aws.amazon.com/AmazonS3/latest/API/API_CSVInput.html#AmazonS3-Type-CSVInput-QuoteCharacter
            df = self.read_csv_with_serializer(
                csv_file_path=input_file_path,
                csv_input_serializer=input_serializer,
                compression_type=compression_type,
            )
        elif isinstance(input_serializer, s3_select.ParquetInputSerializer):
            parquet_file = parquet.ParquetFile(input_file_path)
            df = parquet_file.read().to_pandas()

        filtered_df = self.query_dataframe_with_sql(
            raw_sql=raw_sql,
            df=df,
        )

        if isinstance(output_serializer, s3_select.JSONOutputSerializer):
            if output_serializer.RecordDelimiter != "\n":
                raise NotImplementedError(
                    "Only newline ('\n') is supported as the record delimiter for JSON output in localdev"
                )
            result = filtered_df.to_json(
                orient="records", lines=True, date_format="iso"
            )
        elif isinstance(output_serializer, s3_select.CSVOutputSerializer):
            result = self.output_csv_with_serializer(
                df=filtered_df,
                output_serializer=output_serializer,
            )

        yield result

    def query_dataframe_with_sql(
        self,
        raw_sql: str,
        df: pd.DataFrame,
    ) -> pd.DataFrame:
        # s3 select requires the from clause to use the identifier "s3object"
        # it is case insensitive however so people's queries may use different cases
        S3_OBJECT_QUERY_IDENTIFIER = "s3object"

        con = duckdb.connect(database=":memory:")
        pattern = re.compile(rf"{S3_OBJECT_QUERY_IDENTIFIER}", re.IGNORECASE)
        raw_sql = pattern.sub(S3_OBJECT_QUERY_IDENTIFIER, raw_sql)
        con.register(S3_OBJECT_QUERY_IDENTIFIER, df)

        result = con.execute(raw_sql)

        df_filtered = result.fetchdf()

        return df_filtered

    def read_csv_with_serializer(
        self,
        csv_file_path: str,
        csv_input_serializer: s3_select.CSVInputSerializer,
        compression_type: s3_select.CompressionType | None = None,
    ) -> pd.DataFrame:
        input_serializer = csv_input_serializer.to_dict()

        field_delimiter = input_serializer.get("FieldDelimiter", ",")
        compression = compression_type.lower() if compression_type else "infer"
        quotechar = input_serializer.get("QuoteCharacter", '"')
        escapechar = input_serializer.get("QuoteEscapeCharacter", "\\")
        comment = input_serializer.get("Comments")
        record_delimiter = input_serializer.get("RecordDelimiter")
        header: int | None | str

        if "FileHeaderInfo" in input_serializer.keys():
            if input_serializer["FileHeaderInfo"] == "NONE":
                header = None
            elif input_serializer["FileHeaderInfo"] == "IGNORE":
                raise NotImplementedError(
                    "Value `IGNORE` of column selection not implemented for localdev"
                )
            elif input_serializer["FileHeaderInfo"] == "USE":
                header = 0
            else:
                header = "infer"
        else:
            header = "infer"

        # Define default values for each parameter
        default_kwargs = {
            "sep": field_delimiter,
            "compression": compression,
            "quotechar": quotechar,
            "escapechar": escapechar,
            "comment": comment,
            "header": header,
            "lineterminator": record_delimiter if record_delimiter != "\n" else None,
            "engine": "c" if record_delimiter != "\n" else None,
        }

        kwargs = {
            key: value
            for key, value in default_kwargs.items()
            if value is not None or key == "header"
        }

        df = pd.read_csv(filepath_or_buffer=csv_file_path, **kwargs)

        return df

    def output_csv_with_serializer(
        self,
        df: pd.DataFrame,
        output_serializer: s3_select.CSVOutputSerializer,
    ) -> str:
        output_serializer_dict = output_serializer.to_dict()

        field_delimiter = output_serializer_dict.get("FieldDelimiter", ",")
        quotechar = output_serializer_dict.get("QuoteCharacter", '"')
        escapechar = output_serializer_dict.get("QuoteEscapeCharacter", "\\")
        record_delimiter = output_serializer_dict.get("RecordDelimiter", "\n")

        if "QuoteFields" in output_serializer_dict.keys():
            if output_serializer_dict["QuoteFields"] == "ALWAYS":
                quoting = csv.QUOTE_ALL
            else:
                quoting = None
        else:
            quoting = None

        default_kwargs = {
            "sep": field_delimiter,
            "lineterminator": record_delimiter if record_delimiter != "\n" else None,
            "quotechar": quotechar,
            "escapechar": escapechar,
            "quoting": quoting,
        }

        kwargs = {
            key: value for key, value in default_kwargs.items() if value is not None
        }

        result = df.to_csv(index=False, **kwargs)

        return result

    def fetch_url(
        self,
        key_path: str,
        expires_in: int = 60,
        response_headers: dict[str, str] | None = None,
        version_id: str | None = None,
    ) -> str:
        # N.B. This is a bit brittle, and will only work if the following things are also in place:
        # - settings.MEDIA_ROOT and settings.MEDIA_URL need to be set.
        # - Django must be configured to serve locally uploaded file as per these instructions:
        #   https://docs.djangoproject.com/en/4.2/howto/static-files/#serving-files-uploaded-by-a-user-during-development
        # - A custom storage_root is not passed to the LocalFileStore constructor.
        if key_path.startswith(self.storage_root):
            url_path = key_path.split(self.storage_root)[1]
            # Remove the bucket name from the url path, because it's already going to be added
            # to the fetch URL.
            url_path = url_path.replace(self.bucket_name, "")
        else:
            # This won't give a working url, but at least it won't raise an exception.
            url_path = key_path

        if version_id:
            path_components = os.path.split(url_path)
            url_path = os.path.join(path_components[0], version_id, path_components[1])

        return f"{settings.MEDIA_URL}{self.bucket_name}/{url_path}"

    def generate_presigned_post(
        self, *, key_path: str, expires_in: int = 60
    ) -> PreSignedPost:
        return PreSignedPost(
            # Resolves to a localdev/storage url
            url=reverse("fake-presigned-post-upload"),
            key=key_path,
            # These fields should be passed in the body of the upload POST request by clients that wish to upload a file
            # In production, they contain the signature used by AWS to verify the upload.
            # Locally, however, we return the bucket name and key path so that when the upload request is made
            # we know where to store the file.
            fields={"key_path": key_path, "bucket_name": self.bucket_name},
        )

    def exists(self, key_path: str, as_file: bool = False) -> bool:
        file_path = os.path.join(self.storage_root, self.bucket_name, key_path)
        if as_file:
            return os.path.isfile(file_path)
        else:
            return os.path.exists(file_path)

    def list_s3_keys(self, namespace: str = "") -> Iterable[S3Object]:
        for filename in self.list_files(namespace):
            yield S3Object(bucket_name=self.bucket_name, key=filename)

    def list_files(self, namespace: str = "") -> Iterable[str]:
        """
        Return a generator that lists all files in the tree below the given directory.

        `namespace`:                   the directory whose contents we want to list .

        Note the following two ways in which the output of this function *differs* from the unix
        `ls` command, in order to mimic the S3 `list` output:

        1. It recurses down through subdirectories.
        2. It returns the namespace with the result.

        So given a directory structure /tmp/my-bucket/a/b/c.pdf, we could expect the following
        session:
        >>>> store = LocalFileStore('my-bucket')
        >>>> store.list_files('a')
        ['a/b/c.pdf']
        """
        prefix = os.path.join(self.storage_root, self.bucket_name, namespace)
        if not self.exists(prefix):
            return []
        else:
            return self._list_files(prefix)

    def get_size_in_bytes(self, *, s3_object: S3Object) -> int:
        filepath = os.path.join(self.storage_root, s3_object.bucket_name, s3_object.key)
        file_stats = os.stat(filepath)
        return file_stats.st_size

    def get_last_modified(self, key_path: str) -> datetime.datetime:
        file_path = os.path.join(self.storage_root, self.bucket_name, key_path)
        if not os.path.exists(file_path):
            raise KeyDoesNotExist(f"Key {key_path} was not found at {file_path}")

        file_stats = os.stat(file_path)
        return datetime.datetime.fromtimestamp(file_stats.st_mtime)

    def copy(self, *, s3_object: S3Object, destination: str) -> S3Object:
        shutil.copyfile(
            src=self._filepath("", s3_object.key), dst=self._filepath("", destination)
        )
        return S3Object(bucket_name=self.bucket_name, key=destination)

    def rename(self, *, s3_object: S3Object, destination: str) -> S3Object:
        os.rename(
            src=self._filepath("", s3_object.key), dst=self._filepath("", destination)
        )
        return S3Object(bucket_name=self.bucket_name, key=destination)

    def delete(self, *, s3_object: S3Object) -> None:
        os.remove(self._filepath("", s3_object.key))

    def _filepath(self, namespace: str, filepath: str) -> str:
        key_path = self.make_key_path(namespace=namespace, filepath=filepath)
        filepath = self._filepath_for_key_path(key_path)
        return filepath

    def _filepath_for_key_path(self, key_path: str) -> str:
        filepath = os.path.join(self.storage_root, self.bucket_name, key_path)
        _create_parent_directories(filepath)
        return filepath

    def _list_files(self, root_dir: str) -> Iterable[str]:
        for dirpath, dirnames, filenames in os.walk(root_dir):
            for filename in filenames:
                # Chop off the storage root + bucket_name + /, so that the path we return starts
                # with the namespace. This keeps it compatible with the S3 store. (The empty
                # string as the final argument to `join` ensures we get a trailing slash)
                bucket_prefix = os.path.join(self.storage_root, self.bucket_name, "")
                yield os.path.join(dirpath[len(bucket_prefix) :], filename)

    def _bucket(self) -> LocalStorageBucket:
        return LocalStorageBucket()


class LocalEmailStore(LocalFileStore):
    """
    This is similar to the LocalFileStore class but returns a randomly
    selected HTML email file.
    """

    email_keys = [
        "sample-emails/A-7D673F5A.html",
        "sample-emails/A-111D3A87.html",
        "sample-emails/A-B2250731.html",
        "sample-emails/A-FBC4EF61.html",
    ]

    def __init__(self, bucket_name: str = "", *args: Any, **kwargs: Any) -> None:
        super().__init__(
            bucket_name=bucket_name, storage_root=settings.EMAIL_STORAGE_ROOT
        )

    def fetch_file_contents(
        self, key_path: str, version_id: str | None = None
    ) -> bytes:
        # Randomly select one of the fixture files
        key_path = random.choice(self.email_keys)
        return super().fetch_file_contents(key_path, version_id)


class LocalDocumentStorage(LocalFileStore):
    """
    This is similar to the LocalFileStore class but returns a randomly
    selected document file from the static folder.
    """

    def __init__(self, bucket_name: str, *args: Any, **kwargs: Any) -> None:
        super().__init__(bucket_name, storage_root="")

    def fetch_url(
        self,
        key_path: str,
        expires_in: int = 60,
        response_headers: dict[str, str] | None = None,
        version_id: str | None = None,
    ) -> str:
        bucket = self._bucket()
        return os.path.join(bucket.document_url_base, key_path)

    def list_files(self, namespace: str = "") -> Iterable[str]:
        bucket = self._bucket()
        return bucket.document_filenames


class FileSystemFileStore(LocalFileStore):
    """
    A LocalFileStore which can be constructed using a URI.
    """

    def __init__(self, uri: str) -> None:
        parsed_url = urllib.parse.urlparse(uri)
        if parsed_url.scheme != "file":
            raise ValueError(f"Expected URL starting 'file://'. Got {uri!r}")
        if parsed_url.netloc:
            raise ValueError(f"Expected a file:// URL without a netloc. Got {uri!r}")

        path = parsed_url.path.rstrip("/")
        (storage_root, _, bucket_name) = path.rpartition("/")
        super().__init__(bucket_name, storage_root, use_date_in_key_path=False)


class MemoryFileStore(BaseS3FileStore, Clearable):
    """
    For testing.
    """

    # Use a class-level dict so all instances share the same store.
    buffers: dict[str, dict[str, bytes]] = {}
    versioned_buffers: dict[str, dict[str, dict[str, bytes]]] = defaultdict(
        lambda: defaultdict(lambda: defaultdict(bytes))
    )

    def __init__(self, bucket_name: str, **kwargs: Any) -> None:
        super().__init__(bucket_name, **kwargs)
        if bucket_name not in self.buffers:
            self.buffers[self.bucket_name] = {}

    def store_file(
        self,
        namespace: str,
        filename: str,
        contents: AnyStr | ReadableBinaryFile,
        content_type: str = "",
        overwrite: bool = False,
        metadata: dict[str, str] | None = None,
    ) -> tuple[str, str]:
        key_path = self.make_key_path(namespace=namespace, filepath=filename)
        if key_path not in self.buffers[self.bucket_name] or overwrite:
            self.buffers[self.bucket_name][key_path] = _to_bytes(contents=contents)
        return self.bucket_name, key_path

    def store_versioned_file(
        self,
        key_path: str,
        contents: AnyStr | io.BytesIO,
        content_type: str = "",
    ) -> tuple[str, str, str]:
        version = str(uuid.uuid4())
        self.versioned_buffers[self.bucket_name][key_path][version] = _to_bytes(
            contents=contents
        )
        self.buffers[self.bucket_name][key_path] = _to_bytes(contents=contents)
        return self.bucket_name, key_path, version

    def store_filepath(
        self,
        namespace: str,
        filepath: str,
        overwrite: bool = False,
        dest_filepath: str = "",
    ) -> tuple[str, str]:
        with open(filepath, "rb") as f:
            if not dest_filepath:
                dest_filepath = os.path.basename(filepath)
            return self.store_file(
                namespace, dest_filepath, f.read(), overwrite=overwrite
            )

    def fetch_file_contents(
        self, key_path: str, version_id: str | None = None
    ) -> bytes:
        if version_id:
            versioned_bucket = self.versioned_buffers[self.bucket_name]
            if (
                key_path not in versioned_bucket
                or version_id not in versioned_bucket[key_path]
            ):
                raise KeyDoesNotExist(
                    "Key with path %s and version %s was not found"
                    % key_path
                    % version_id
                )
            return versioned_bucket[key_path][version_id]
        else:
            bucket = self.buffers[self.bucket_name]
            if key_path not in bucket:
                raise KeyDoesNotExist("Key with path %s was not found" % key_path)
            return bucket[key_path]

    def get_key(self, key_path: str, version_id: str | None = None) -> S3Object:
        return S3Object(
            bucket_name=self.bucket_name, key=key_path, version_id=version_id
        )

    def get_file_type(self, key_path: str) -> str:
        mime = magic.Magic(mime=True)
        return mime.from_buffer(self.fetch_file_contents(key_path))

    def fetch_file(
        self, key_path: str, version_id: str | None = None, range: str | None = None
    ) -> StreamingBody:
        raw_stream = io.BytesIO(self.fetch_file_contents(key_path, version_id))
        return StreamingBody(
            raw_stream=raw_stream, content_length=files.size(raw_stream)
        )

    def fetch_url(
        self,
        key_path: str,
        expires_in: int = 60,
        response_headers: dict[str, str] | None = None,
        version_id: str | None = None,
    ) -> str:
        return TEST_FETCH_URL

    def exists(self, key_path: str, as_file: bool = False) -> bool:
        # as_file is defined to match the interface of the other Stores, but is not obviously
        # applicable for keys in this medium
        bucket = self.buffers[self.bucket_name]
        return key_path in bucket

    def list_files(self, namespace: str = "") -> Iterable[str]:
        bucket = self.buffers[self.bucket_name]
        if not namespace:
            return bucket.keys()
        prefix = namespace
        if not prefix.endswith("/"):
            prefix += "/"

        # Using a list comprehension here makes a copy of the keys, which stops a RuntimeError
        # being thrown if the caller modifies something in the bucket.
        return [key for key in bucket if key.startswith(prefix)]

    def list_s3_keys(self, namespace: str = "") -> Iterable[S3Object]:
        return [self.get_key(path) for path in self.list_files(namespace=namespace)]

    def copy(self, *, s3_object: S3Object, destination: str) -> S3Object:
        bucket = self.buffers[self.bucket_name]
        bucket[destination] = bucket[s3_object.key]
        return S3Object(bucket_name=self.bucket_name, key=destination)

    def rename(self, *, s3_object: S3Object, destination: str) -> S3Object:
        assert s3_object.bucket_name == self.bucket_name
        bucket = self.buffers[self.bucket_name]
        if s3_object.key not in bucket:
            raise KeyDoesNotExist("Key with path %s was not found" % s3_object.key)
        if destination == s3_object.key:
            return s3_object  # No rename required
        bucket[destination] = bucket[s3_object.key]
        del bucket[s3_object.key]
        return S3Object(bucket_name=self.bucket_name, key=destination)

    def delete(self, *, s3_object: S3Object) -> None:
        bucket = self.buffers[self.bucket_name]
        del bucket[s3_object.key]

    def clear(self) -> None:
        """
        Deletes all files from all buckets, across all MemoryFileStore instances.
        """
        for bucket in self.buffers.values():
            bucket.clear()

    def generate_presigned_post(
        self, *, key_path: str, expires_in: int = 60
    ) -> PreSignedPost:
        return PreSignedPost(
            # Resolves to a localdev/storage url
            url=reverse("fake-presigned-post-upload"),
            key=key_path,
            # These fields should be passed in the body of the upload POST request by clients that wish to upload a file
            # In production, they contain the signature used by AWS to verify the upload.
            # Locally, however, we return the bucket name and key path so that when the upload request is made
            # we know where to store the file.
            fields={"key_path": key_path, "bucket_name": self.bucket_name},
        )

    def get_size_in_bytes(self, *, s3_object: S3Object) -> int:
        assert s3_object.bucket_name == self.bucket_name
        return len(self.fetch_file_contents(s3_object.key))

    def get_last_modified(self, key_path: str) -> datetime.datetime:
        """
        Given this is used for testing purposes, it's not as relevant.  Just return current time.
        """
        return datetime.datetime.now()


def store(
    bucket_name: str,
    use_date_in_key_path: bool = True,
    set_acl_bucket_owner: bool = False,
) -> BaseS3FileStore:
    """
    Return the appropriate storage instance for a given bucket.
    """
    storage_class = import_string(settings.STORAGE_BACKEND)
    return storage_class(
        bucket_name=bucket_name,
        use_date_in_key_path=use_date_in_key_path,
        set_acl_bucket_owner=set_acl_bucket_owner,
    )


def email_store(bucket_name: str) -> BaseS3FileStore:
    """
    Return the storage instance for email content.
    """
    store_class = import_string(settings.EMAIL_STORAGE_BACKEND)
    return store_class(bucket_name)


def fileserver_store(use_date_in_key_path: bool = True) -> BaseS3FileStore:
    return store(settings.S3_FILESERVER_BUCKET, use_date_in_key_path)


def flows_outbound_store() -> BaseS3FileStore:
    return store(settings.S3_PRODUCTION_FLOWS_OUTBOUND)


def user_documents(use_date_in_key_path: bool = True) -> BaseS3FileStore:
    """
    Return the user documents store.
    """
    return store(
        settings.S3_USER_DOCUMENTS_BUCKET, use_date_in_key_path=use_date_in_key_path
    )


def archive(use_date_in_key_path: bool = True) -> BaseS3FileStore:
    """
    Return the file archive store.
    """
    return store(settings.S3_ARCHIVE_BUCKET, use_date_in_key_path)


def support_documents_store() -> BaseS3FileStore:
    """
    Return the complaint related documents store.
    """
    store_class = import_string(settings.DOCUMENT_STORAGE_BACKEND)
    return store_class(settings.S3_SUPPORT_DOCUMENTS_BUCKET)


def voice_audio_statics_store() -> BaseS3FileStore:
    return store(settings.S3_VOICE_AUDIO_BUCKET, use_date_in_key_path=False)


def store_file_attachment(attachment: BinaryIO, namespace: str) -> tuple[str, str]:
    attachment_store = support_documents_store()
    contents = attachment.read()
    contents_hash = base64.urlsafe_b64encode(
        hashlib.blake2b(contents, digest_size=33).digest()
    ).decode("utf8")
    s3_bucket, s3_key = attachment_store.store_file(
        namespace=namespace,
        filename=os.path.join(contents_hash, attachment.name),
        contents=contents,
        overwrite=True,
    )
    return s3_bucket, s3_key


def outbound_flow_store() -> BaseS3FileStore:
    storage_class = import_string(settings.STORAGE_BACKEND)
    if storage_class == S3FileStore:
        return storage_class(
            bucket_name=settings.INTEGRATION_FLOW_S3_OUTBOUND_BUCKET,
            use_date_in_key_path=False,
        )
    else:
        return storage_class(bucket_name=settings.INTEGRATION_FLOW_S3_OUTBOUND_BUCKET)


def from_uri(
    uri: str,
) -> FileSystemFileStore | S3SubdirectoryFileStore | MemoryFileStore:
    """
    :raises ValueError: if the URI does not contain a scheme for a supported storage system.
    """
    scheme = urllib.parse.urlparse(uri).scheme
    if scheme == "file":
        return FileSystemFileStore(uri)
    elif scheme == "s3":
        return S3SubdirectoryFileStore(uri)
    elif scheme == "memory":
        return MemoryFileStore(uri)
    else:
        raise ValueError(f"Expected a URL for a supported storage system. Got {uri!r}")


def user_media_store() -> BaseS3FileStore:
    """
    Returns the store where Kraken users can upload media such as audio files.
    """
    return store(settings.S3_USER_MEDIA_BUCKET)


def line_file_store() -> BaseS3FileStore:
    """
    Returns the store for inbound attachments from LINE (Ink).
    """
    return store(settings.LINE_INBOUND_ATTACHMENTS_BUCKET)


# Private


def _create_parent_directories(filepath: str) -> None:
    os.makedirs(os.path.dirname(filepath), mode=0o755, exist_ok=True)


def _to_stream(*, contents: AnyStr | ReadableBinaryFile) -> ReadableBinaryFile:
    """
    Return the given object expressed as an IO stream object.
    """
    if isinstance(contents, str):
        return io.BytesIO(contents.encode())
    elif isinstance(contents, bytes):
        return io.BytesIO(contents)
    return contents


def _to_bytes(*, contents: AnyStr | ReadableBinaryFile) -> bytes:
    """
    Return the given object expressed as a bytes object.
    """
    if isinstance(contents, str):
        return contents.encode()
    elif isinstance(contents, bytes):
        return contents
    return contents.read()



import datetime
import typing

import pytz

from . import exceptions


__all__ = [
    "convert_sp_and_date_to_utc",
    "convert_utc_to_sp_and_date",
    "number_of_periods_in_timedelta",
]


# Time zones implemented
UTC_TZ = "UTC"
GB_TZ = "Europe/London"


def _to_timezone(local_time: datetime.datetime, timezone_str: str) -> datetime.datetime:
    """
    Converts an aware datetime to another time zone.
    """
    timezone = pytz.timezone(timezone_str)
    # Rebasing to local time zone
    rebased_local_time = local_time.astimezone(timezone)
    # Normalizing for daylight saving rules
    return timezone.normalize(rebased_local_time)


def _get_first_delivery_time(
    date: datetime.date, timezone_str: str, is_wholesale: bool
) -> datetime.datetime:
    """
    Return an aware datetime for the start of the first settlement period on a given date.
    """
    midnight = datetime.datetime(date.year, date.month, date.day)
    offset_midnight = midnight - datetime.timedelta(hours=1)
    # First settlement period
    cases = {
        # UTC: first delivery starts at 00:00:00 UTC on the day
        UTC_TZ: midnight,
        # GB retail: first delivery starts at 00:00:00 Europe/London on the day
        # GB wholesale: first delivery starts at 23:00:00 Europe/London on the previous day
        GB_TZ: offset_midnight if is_wholesale else midnight,
    }
    # Time zone aware
    try:
        return pytz.timezone(timezone_str).localize(cases[timezone_str])
    except AttributeError:
        raise exceptions.SettlementPeriodError("Time zone not implemented")


def _get_delivery_date(
    local_time: datetime.datetime, timezone_str: str, is_wholesale: bool
) -> datetime.date:
    """
    Return the date of the settlement period relative to a tz-aware datetime.
    """
    period_date = local_time.date()
    offset_period_date = period_date
    if local_time.hour >= 23:
        offset_period_date += datetime.timedelta(days=1)
    # Date of the first settlement period
    cases = {
        # UTC: first delivery starts at 00:00:00 UTC on the day
        UTC_TZ: period_date,
        # GB retail: first delivery starts at 00:00:00 Europe/London on the day
        # GB wholesale: first delivery starts at 23:00:00 Europe/London on the previous day
        GB_TZ: offset_period_date if is_wholesale else period_date,
    }
    # Time zone aware
    try:
        return cases[timezone_str]
    except AttributeError:
        raise exceptions.SettlementPeriodError("Time zone not implemented")


def _round_local_down_to_hh(local_time: datetime.datetime) -> datetime.datetime:
    if local_time.minute < 30:
        return local_time - datetime.timedelta(minutes=local_time.minute)
    else:
        return local_time - datetime.timedelta(minutes=local_time.minute - 30)


def convert_sp_and_date_to_local(
    sp: int, date: datetime.date, timezone_str: str, is_wholesale: bool
) -> datetime.datetime:
    """
    Return an aware datetime for the start of a given settlement period.
    """
    if sp not in range(1, 51):
        raise exceptions.SettlementPeriodError("Settlement period not valid")
    # First settlement period in the time zone
    first_period_start = _get_first_delivery_time(date, timezone_str, is_wholesale)
    # Start of the settlement period in local time
    local_time = first_period_start + datetime.timedelta(minutes=30 * (sp - 1))
    # Normalizing for daylight saving rules
    return pytz.timezone(timezone_str).normalize(local_time)


def convert_sp_and_date_to_utc(
    sp: int, date: datetime.date, timezone_str: str = GB_TZ, is_wholesale: bool = False
) -> datetime.datetime:
    """
    Return an UTC-aware datetime for the start of a given settlement period.
    """
    local_time = convert_sp_and_date_to_local(sp, date, timezone_str, is_wholesale)
    return _to_timezone(local_time, UTC_TZ).replace(tzinfo=datetime.timezone.utc)


def convert_local_to_sp_and_date(
    local_time: datetime.datetime, is_wholesale: bool = False
) -> typing.Tuple[int, datetime.date]:
    """
    Return the date and settlement period from a given tz-aware datetime.
    """
    try:
        timezone_str = str(local_time.tzinfo)
    except AttributeError:
        raise exceptions.SettlementPeriodError("Not a tz-aware datetime")
    # Round to the nearest half hour
    half_hourly_time = _round_local_down_to_hh(local_time)
    # Date of the settlement period in the time zone
    delivery_date = _get_delivery_date(half_hourly_time, timezone_str, is_wholesale)
    # First settlement period in the time zone
    first_delivery_time = _get_first_delivery_time(
        delivery_date, timezone_str, is_wholesale
    )
    # Fetch settlement period
    delta = half_hourly_time - first_delivery_time
    settlement_period = ((int(delta.total_seconds()) // 60) + 30) // 30
    return settlement_period, delivery_date


def convert_utc_to_sp_and_date(
    utc_time: datetime.datetime, timezone_str: str = GB_TZ, is_wholesale: bool = False
) -> typing.Tuple[int, datetime.date]:
    """
    Return the local date and settlement period from a given UTC-aware datetime.
    """
    local_time = _to_timezone(utc_time, timezone_str)
    return convert_local_to_sp_and_date(local_time, is_wholesale)


def number_of_periods_in_timedelta(delta: datetime.timedelta) -> float:
    """
    Return the number of half-hourly settlement periods in the given timedelta
    """
    return (delta.total_seconds() / 60) / 30



from __future__ import annotations

import dataclasses
import json
import os
import subprocess
from typing import Any

import pact
import requests


@dataclasses.dataclass(frozen=True)  # Use kw_only=True once we're on Python 3.10.
class PactOptions:
    broker_url: str
    broker_username: str
    broker_password: str
    consumer_name: str
    provider_name: str
    consumer_version: str
    log_path: str


class PactConsumerClient:
    def __init__(self, base_url: str) -> None:
        self.base_url = base_url

    def post(
        self,
        path: str,
        data: dict[str, Any],
        token: str | None = None,
    ) -> dict[str, Any]:
        url = self.base_url + path
        headers = {"Content-Type": "application/json"}
        if token:
            headers["Authorization"] = token
        response = requests.post(
            url, data=json.dumps(data), headers=headers, verify=False
        )
        return response.json()


def pact_service(*, options: PactOptions, publish_to_broker: bool) -> pact.Pact:
    service = pact.Consumer(
        name=options.consumer_name,
        tag_with_git_branch=True,
        version=options.consumer_version,
    ).has_pact_with(
        pact.Provider(options.provider_name),
        publish_to_broker=publish_to_broker,
        broker_base_url=options.broker_url,
        broker_username=options.broker_username,
        broker_password=options.broker_password,
        pact_dir=options.log_path,
        log_dir=options.log_path,
    )

    return service


def get_unique_version_hash() -> str | None:
    """
    Get a unique version number for the pact verification for the current git revision.

    - Get git hash for the current git revision.

    """
    if "CIRCLECI" not in os.environ:
        return None
    version = _git_revision_hash()
    return version


def _git_revision_hash() -> str:
    """
    Return the git revision hash for the current branch.
    """

    git_hash = subprocess.check_output(["git", "rev-parse", "HEAD"])

    return git_hash.decode("utf8").rstrip()


def get_git_branch_name() -> str | None:
    """
    Get the current git branch name.
    """
    branch_name = _git_branch_name()

    if "fatal: not a git repository" in branch_name:
        return None

    return branch_name


def _git_branch_name() -> str:
    """
    This runs git rev-parse --abbrev-ref HEAD to get the current branch name.
    """
    git_branch = subprocess.check_output(
        [
            "git",
            "rev-parse",
            "--abbrev-ref",
            "HEAD",
        ]
    )

    return git_branch.decode("utf8").rstrip()



from typing import Generic, Tuple, TypeVar, Union

from django.contrib.auth import models as auth_models
from django.db import models
from django.db.models.expressions import Combinable
from django.http import HttpRequest


# A type variable which can be used in generic types, and represents a Django model of some
# description
Model = TypeVar("Model", bound=models.Model)


# A type variable for representing arbitrary user models
User = TypeVar("User", bound=auth_models.AbstractBaseUser)


# Helpers for declaring django relations on classes
# These are one-item Unions so that mypy knows they are type aliases and not strings
ForeignKey = Union["models.ForeignKey[Union[Model, Combinable], Model]"]
OptionalForeignKey = Union[
    "models.ForeignKey[Union[Model, Combinable, None], Union[Model, None]]"
]

OneToOneField = Union["models.OneToOneField[Union[Model, Combinable], Model]"]
OptionalOneToOneField = Union[
    "models.OneToOneField[Union[Model, Combinable, None], Union[Model, None]]"
]

# A type variable to describe the Django model choices kwarg
Choices = Tuple[Tuple[str, str], ...]


class AuthenticatedRequest(HttpRequest, Generic[User]):
    """
    This class represents HttpRequests that are guaranteed to have an auth user associated
    with them, for use when hinting requests that are known to be authenticated.
    """

    user: User



from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.loader import MigrationLoader


def check_migrations() -> bool:
    """
    Check if there are any migrations that haven't been applied yet.
    """
    return _num_unapplied_migrations() == 0


def _num_unapplied_migrations() -> int:
    """
    Return the number of unapplied migrations.
    """
    connection = connections[DEFAULT_DB_ALIAS]
    loader = MigrationLoader(connection)
    graph = loader.graph

    # Count unapplied migrations
    num_unapplied_migrations = 0
    for app_name in loader.migrated_apps:
        for node in graph.leaf_nodes(app_name):
            for plan_node in graph.forwards_plan(node):
                if plan_node not in loader.applied_migrations:
                    num_unapplied_migrations += 1

    return num_unapplied_migrations
